#!/usr/bin/env python3

import torch
from torch.utils.data import Dataset, Subset
import numpy as np

from typing import Sequence, overload


class TensorListDataset(Dataset):
    r"""The dataset class that has a :any:`torch.Tensor` as inputs
    and :any:`list`\[:any:`int`\] as labels.
    It inherits :any:`torch.utils.data.Dataset`.

    Args:
        data (torch.Tensor): The inputs.
        targets (list[int]): The labels.
        **kwargs: Keyword arguments passed to
            :any:`torch.utils.data.Dataset`.

    :Example:
        >>> import torch
        >>> from trojanzoo.utils.data import TensorListDataset
        >>>
        >>> data = torch.ones(10, 3, 32, 32)
        >>> targets = list(range(10))
        >>> dataset = TensorListDataset(data, targets)
        >>> x, y = dataset[3]
        >>> x.shape
        torch.Size([3, 32, 32])
        >>> y
        3
    """

    def __init__(self, data: torch.Tensor = None,
                 targets: list[int] = None, **kwargs):
        super().__init__(**kwargs)
        self.data = data
        self.targets = targets
        assert len(self.data) == len(self.targets)
        self.__length = len(self.targets)

    @overload
    def __getitem__(self, index: int) -> tuple[torch.Tensor, int]:
        ...

    @overload
    def __getitem__(self, index: slice) -> tuple[torch.Tensor, list[int]]:
        ...

    def __getitem__(self, index):
        return self.data[index], self.targets[index]

    def __len__(self) -> int:
        return self.__length


def dataset_to_tensor(dataset: Dataset) -> tuple[torch.Tensor, torch.Tensor]:
    r"""Transform a :any:`torch.utils.data.Dataset` to ``(data, targets)`` tensor tuple
    by traversing all elements.

    Args:
        dataset (torch.utils.data.Dataset): The dataset.

    Returns:
        (torch.Tensor, torch.Tensor): The tuple of ``(data, targets)``.

    :Example:
        >>> from torchvision.datasets import MNIST
        >>> import torchvision.transforms as transforms
        >>> from trojanzoo.utils.data import dataset_to_tensor
        >>>
        >>> transform = transforms.Compose([
            transforms.PILToTensor(),
            transforms.ConvertImageDtype(torch.float)])
        >>> dataset = MNIST('./', train=False, download=True,
                            transform=transform)
        >>> data, targets = dataset_to_tensor(dataset)
        >>> data.shape
        torch.Size([10000, 1, 28, 28])
        >>> targets.shape
        torch.Size([10000])
        >>> targets.dtype
        torch.int64
    """
    data, targets = list(zip(*dataset))[:2]
    return torch.stack(data), torch.as_tensor(targets, dtype=torch.long)


def sample_batch(dataset: Dataset, batch_size: int = None,
                 idx: Sequence[int] = []) -> tuple[torch.Tensor, torch.Tensor]:
    r"""Sample a batch from dataset by calling

    .. parsed-literal::
        :func:`dataset_to_tensor`\(:any:`torch.utils.data.Subset`\(dataset, idx))

    Args:
        dataset (torch.utils.data.Dataset): The dataset to sample.
        batch_size (int): The batch size to sample
            when :attr:`idx` is ``None``.
            Defaults to ``None``.
        idx (Sequence[int]): The index list of each sample in dataset.
            If empty, randomly sample a batch with given :attr:`batch_size`.
            Defaults to ``[]``.

    Returns:
        (torch.Tensor, torch.Tensor): The tuple of sampled batch ``(data, targets)``.

    :Example:
        >>> import torch
        >>> from trojanzoo.utils.data import TensorListDataset, sample_batch
        >>>
        >>> data = torch.ones(10, 3, 32, 32)
        >>> targets = list(range(10))
        >>> dataset = TensorListDataset(data, targets)
        >>> x, y = sample_batch(dataset, idx=[1, 2])
        >>> x.shape
        torch.Size([2, 3, 32, 32])
        >>> y
        tensor([1, 2])
        >>> x, y = sample_batch(dataset, batch_size=4)
        >>> y
        tensor([6, 3, 2, 5])
    """
    if len(idx) == 0:
        idx = torch.randperm(len(dataset))[:batch_size]
    else:
        idx = torch.as_tensor(idx, dtype=torch.int, device='cpu')
    return dataset_to_tensor(Subset(dataset, idx))


def split_dataset(dataset: Dataset | Subset,
                  length: int = None, percent: float = None,
                  shuffle: bool = True, seed: int = None
                  ) -> tuple[Subset, Subset]:
    r"""Split a dataset into two subsets.

    Args:
        dataset (torch.utils.data.Dataset): The dataset to split.
        length (int): The length of the first subset.
            This argument cannot be used together with :attr:`percent`.
            If ``None``, use :attr:`percent` to calculate length instead.
            Defaults to ``None``.
        percent (float): The split ratio for the first subset.
            This argument cannot be used together with :attr:`length`.
            ``length = percent * len(dataset)``.
            Defaults to ``None``.
        shuffle (bool): Whether to shuffle the dataset.
            Defaults to ``True``.
        seed (bool): The random seed to split dataset
            using :any:`numpy.random.shuffle`.
            Defaults to ``None``.

    Returns:
        (torch.utils.data.Subset, torch.utils.data.Subset):
            The two splitted subsets.

    :Example:
        >>> import torch
        >>> from trojanzoo.utils.data import TensorListDataset, split_dataset
        >>>
        >>> data = torch.ones(11, 3, 32, 32)
        >>> targets = list(range(11))
        >>> dataset = TensorListDataset(data, targets)
        >>> set1, set2 = split_dataset(dataset, length=3)
        >>> len(set1), len(set2)
        (3, 8)
        >>> set3, set4 = split_dataset(dataset, percent=0.5)
        >>> len(set3), len(set4)
        (5, 6)

    Note:
        This is the implementation of :meth:`trojanzoo.datasets.Dataset.split_dataset`.
        The difference is that this method will NOT set :attr:`seed`
        as ``env['data_seed']`` when it is ``None``.
    """
    assert (length is None) != (percent is None)  # XOR check
    length = length if length is not None else int(len(dataset) * percent)
    indices = np.arange(len(dataset))
    if shuffle:
        if seed is not None:
            np.random.seed(seed)
        np.random.shuffle(indices)
    if isinstance(dataset, Subset):
        idx = np.array(dataset.indices)
        indices = idx[indices]
        dataset = dataset.dataset
    subset1 = Subset(dataset, indices[:length])
    subset2 = Subset(dataset, indices[length:])
    return subset1, subset2


def get_class_subset(dataset: Dataset,
                     class_list: int | list[int]
                     ) -> Subset:
    r"""Get a subset from dataset with certain classes.

    Args:
        dataset (torch.utils.data.Dataset): The entire dataset.
        class_list (int | list[int]): The class list to pick.

    Returns:
        torch.utils.data.Subset: The subset with labels in :attr:`class_list`.

    :Example:
        >>> import torch
        >>> from trojanzoo.utils.data import get_class_subset, TensorListDataset
        >>>
        >>> data = torch.ones(11, 3, 32, 32)
        >>> targets = list(range(11))
        >>> dataset = TensorListDataset(data, targets)
        >>> subset = get_class_subset(dataset, class_list=[2, 3])
        >>> len(subset)
        2
    """
    class_list = [class_list] if isinstance(class_list, int) else class_list
    indices = np.arange(len(dataset))
    while isinstance(dataset, Subset):
        idx = np.array(dataset.indices)
        indices = idx[indices]
        dataset = dataset.dataset
    _, targets = dataset_to_tensor(dataset=dataset)
    idx_bool = np.isin(targets.numpy(), class_list)
    idx = np.arange(len(dataset))[idx_bool]
    idx = np.intersect1d(idx, indices)
    return Subset(dataset, idx)



#!/usr/bin/env python3

from ...abstract import InputFiltering
from trojanzoo.environ import env
from trojanzoo.utils.data import TensorListDataset
from trojanzoo.utils.logger import MetricLogger, SmoothedValue
from trojanzoo.utils.metric import mask_jaccard

import torch
import numpy as np
from sklearn.cluster import KMeans

import argparse


class Neo(InputFiltering):
    r"""Neo proposed by Sakshi Udeshi
    from Singapore University of Technology and Design
    in 2019.

    It is a input filtering backdoor defense
    that inherits :class:`trojanvision.defenses.InputFiltering`.

    The defense procedure is:

    - For a test input, Neo generates its different variants
      with a random region masked by the input's dominant color
      using :any:`sklearn.cluster.KMeans`.
    - For each variant, if its classification is different,
      check if the pixels from masked region is a trigger
      by evaluating its ASR.
    - If ASR of any variant exceeds the :attr:`neo_asr_threshold`,
      the test input is regarded as poisoned.

    See Also:
        * paper: `Model Agnostic Defence against Backdoor Attacks in Machine Learning`_
        * code: https://github.com/sakshiudeshi/Neo

    Note:
        Neo assumes the defender has the knowledge of the trigger size.

    Args:
        neo_asr_threshold (float): ASR threshold.
            Defaults to ``0.8``.
        neo_kmeans_num (int): Number of KMean clusters.
            Defaults to ``3``.
        neo_sample_num (int): Number of sampled masked regions.
            Defaults to ``100``.

    Attributes:
        mark_size(tuple[int, int]): Watermark size ``(h, w)`` of ``self.attack.mark``.

    .. _Model Agnostic Defence against Backdoor Attacks in Machine Learning:
        https://arxiv.org/abs/1908.02203
    """
    name: str = 'neo'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--neo_asr_threshold', type=float,
                           help='ASR threshold for neo defense '
                           '(default: 0.8)')
        group.add_argument('--neo_kmeans_num', type=int,
                           help='number of k-mean clusters for neo defense '
                           '(default: 3)')
        group.add_argument('--neo_sample_num', type=int,
                           help='number of sampled masked regions '
                           '(default: 100)')
        return group

    def __init__(self, neo_asr_threshold: float = 0.8, neo_kmeans_num: int = 3,
                 neo_sample_num: int = 100, **kwargs):
        super().__init__(**kwargs)
        self.param_list['neo'] = ['neo_asr_threshold', 'neo_kmeans_num', 'neo_sample_num']
        self.neo_asr_threshold = neo_asr_threshold
        self.neo_kmeans_num = neo_kmeans_num
        self.neo_sample_num = neo_sample_num

        self.mark_size: tuple[int, int] = [self.attack.mark.mark_height, self.attack.mark.mark_width]
        self.select_num = self.mark_size[0] * self.mark_size[1]

    @torch.no_grad()
    def get_pred_labels(self) -> torch.Tensor:
        logger = MetricLogger(meter_length=40)
        str_format = '{global_avg:5.3f} ({min:5.3f}, {max:5.3f})'
        logger.create_meters(cls_diff=str_format, jaccard_idx=str_format)
        test_set = TensorListDataset(self.test_input, self.test_label)
        test_loader = self.dataset.get_dataloader(mode='valid', dataset=test_set, batch_size=1)
        clean_list = []
        poison_list = []
        for data in logger.log_every(test_loader):
            _input: torch.Tensor = data[0]
            _input = _input.to(env['device'], non_blocking=True)
            trigger_input = self.attack.add_mark(_input)
            clean_list.append(self.get_pred_label(_input[0], logger=logger))
            poison_list.append(self.get_pred_label(trigger_input[0], logger=logger))
        return torch.as_tensor(clean_list + poison_list, dtype=torch.bool)

    def get_pred_label(self, img: torch.Tensor, logger: MetricLogger = None) -> bool:
        r"""Get the prediction label of one certain image (poisoned or not).

        Args:
            img (torch.Tensor): Image tensor (on GPU) with shape ``(C, H, W)``.
            logger (trojanzoo.utils.logger.MetricLogger):
                output logger.
                Defaults to ``None``.

        Returns:
            bool: Whether the image tensor :attr:`img` is poisoned.
        """
        # get dominant color
        dom_c = self.get_dominant_color(img).unsqueeze(-1).unsqueeze(-1)  # (C, 1, 1)

        # generate random numbers
        height, width = img.shape[-2:]
        pos_height = torch.randint(
            low=0, high=height - self.mark_size[0], size=[self.neo_sample_num, 1])
        pos_width = torch.randint(
            low=0, high=width - self.mark_size[1], size=[self.neo_sample_num, 1])
        pos_list = torch.stack([pos_height, pos_width], dim=1)    # (neo_sample_num, 2)
        # block potential triggers on _input
        block_input = img.repeat(self.neo_sample_num, 1, 1, 1)  # (neo_sample_num, C, H, W)
        for i in range(self.neo_sample_num):
            x = pos_list[i][0]
            y = pos_list[i][1]
            block_input[i, :, x:x + self.mark_size[0], y:y + self.mark_size[1]] = dom_c
        # get potential triggers
        org_class = self.model.get_class(img.unsqueeze(0)).item()   # (1)
        block_class = self.model.get_class(block_input).cpu()   # (neo_sample_num)

        # confirm triggers
        pos_pairs = pos_list[block_class != org_class]   # (*, 2)
        for pos in pos_pairs:
            self.attack.mark.mark_height_offset = pos[0]
            self.attack.mark.mark_width_offset = pos[1]
            self.attack.mark.mark.fill_(1.0)
            self.attack.mark.mark[:-1] = img[..., pos[0]:pos[0] + self.mark_size[0],
                                             pos[1]:pos[1] + self.mark_size[1]]
            cls_diff = self.get_cls_diff()
            if cls_diff > self.neo_asr_threshold:
                jaccard_idx = mask_jaccard(self.attack.mark.get_mask(),
                                           self.real_mask,
                                           select_num=self.select_num)
                logger.update(cls_diff=cls_diff, jaccard_idx=jaccard_idx)
                return True
        return False

    def get_dominant_color(self, img: torch.Tensor) -> torch.Tensor:
        r"""Get dominant color for one image tensor
        using :class:`sklearn.cluster.KMeans`.

        Args:
            img (torch.Tensor): Image tensor with shape ``(C, H, W)``.

        Returns:
            torch.Tensor: Dominant color tensor with shape ``(C)``.
        """
        img_np: np.ndarray = img.flatten(1).transpose(0, 1).cpu().numpy()    # (*, C)
        kmeans_result = KMeans(n_clusters=self.neo_kmeans_num).fit(img_np)
        unique, counts = np.unique(kmeans_result.labels_, return_counts=True)
        center = kmeans_result.cluster_centers_[unique[np.argmax(counts)]]
        return torch.as_tensor(center)

    def get_cls_diff(self):
        r"""Get classification difference between
        original inputs and trigger inputs.

        Returns:
            float: Classification difference percentage.
        """
        diff = SmoothedValue()
        for data in self.dataset.loader['valid']:
            _input, _ = self.model.get_data(data)
            _class = self.model.get_class(_input)
            trigger_input = self.attack.add_mark(_input)
            trigger_class = self.model.get_class(trigger_input)
            result = _class.not_equal(trigger_class)
            diff.update(result.float().mean().item(), len(_input))
        return diff.global_avg



#!/usr/bin/env python3

from .output import prints
from trojanzoo.environ import env

import torch
import torch.cuda


def empty_cache(threshold: float = None):
    r"""Call :any:`torch.cuda.empty_cache` to empty GPU cache when
    :any:`torch.cuda.memory_cached() <torch.cuda.memory_cached>`
    ``>`` :attr:`threshold`\ ``MB``.

    Args:
        threshold (float): The cached memory threshold (MB).
            Defaults to ``env['cache_threshold']``.
    """
    threshold = threshold if threshold is not None else env['cache_threshold']
    if threshold is not None and env['num_gpus']:
        if torch.cuda.memory_cached() > threshold * (1 << 20):
            torch.cuda.empty_cache()


def output_memory(device: None | str | int | torch.device = None,
                  full: bool = False, indent: int = 0, **kwargs):
    r"""Output memory usage information.

    Args:
        device (None | str | int | ~torch.torch.device):
            Passed to :any:`torch.cuda.memory_summary`
            or :any:`torch.cuda.memory_allocated` |
            :any:`torch.cuda.memory_reserved`.
            Defaults to ``None``.
        full (bool):
            Whether to call :any:`torch.cuda.memory_summary`.
            Otherwise, call :any:`torch.cuda.memory_allocated`
            and :any:`torch.cuda.memory_reserved`.
            Defaults to ``False``.
        indent (int): The space indent for the entire string.
            Defaults to ``0``.
        **kwargs: Keyword arguments passed to
            :any:`torch.cuda.memory_summary`.
    """
    if full:
        prints(torch.cuda.memory_summary(device=device, **kwargs))
    else:
        prints('memory allocated: '.ljust(20),
               bytes2size(torch.cuda.memory_allocated(device=device)),
               indent=indent)
        prints('memory reserved: '.ljust(20),
               bytes2size(torch.cuda.memory_reserved(device=device)),
               indent=indent)


def bytes2size(_bytes: int) -> str:
    if _bytes < 2 << 10:
        return '%d bytes' % _bytes
    elif _bytes < 2 << 20:
        return '%.3f KB' % (float(_bytes) / (1 << 10))
    elif _bytes < 2 << 30:
        return '%.3f MB' % (float(_bytes) / (1 << 20))
    else:
        return '%.3f GB' % (float(_bytes) / (1 << 30))



#!/usr/bin/env python3

r"""
CUDA_VISIBLE_DEVICES=0 python examples/backdoor_attack.py --color --verbose 1 --pretrained --validate_interval 1 --epochs 20 --lr 0.01 --attack refool --tqdm --efficient
"""  # noqa: E501

from ...abstract import CleanLabelBackdoor
from trojanvision.environ import env
from trojanzoo.utils.data import TensorListDataset, sample_batch
from trojanzoo.utils.logger import MetricLogger
from trojanzoo.utils.output import output_iter

import torch
import torchvision
import torchvision.transforms as transforms
import torchvision.transforms.functional as F
from torchvision.transforms.functional import InterpolationMode

import math
import random
import skimage.metrics
import PIL.Image as Image

import copy
import io
import os
import tarfile
from xml.etree.ElementTree import parse as ET_parse

import argparse
import torch.utils.data

sets: list[tuple[str, str]] = [('2012', 'train'), ('2012', 'val'), ('2007', 'train'), ('2007', 'val'), ('2007', 'test')]
norm = torch.distributions.normal.Normal(loc=0.0, scale=1.0)    # TODO: avoid construction when unused


def read_tensor(fp: str) -> torch.Tensor:
    tensor = F.convert_image_dtype(F.pil_to_tensor(Image.open(fp)))
    return tensor.unsqueeze(0) if tensor.dim() == 2 else tensor


class Refool(CleanLabelBackdoor):
    r"""Reflection Backdoor Attack (Refool) proposed by Yunfei Liu
    from Beihang University in ECCV 2020.

    It inherits :class:`trojanvision.attacks.CleanLabelBackdoor`.

    Note:
        * Trigger size must be the same as image size.
        * Currently, :attr:`mark_alpha` is forced to be ``-1.0``,
          which means to use mean of image and mark to blend them.
          It should be possible to set a manual :attr:`mark_alpha` instead.

    The attack has 3 procedures:

        * **Generate** :attr:`candidate_num` **reflect images from another public dataset (e.g., Pascal VOC) as trigger candidates.**

          - Select a :attr:`reflect class` (e.g., ``'cat'``)
            and a :attr:`background class` (e.g., ``'person'``)
          - Find all images of those 2 classes that
            don't have the object of the other class in them.
          - For image pairs from 2 classes, process and blend them using ``'ghost effect'``
            or ``'focal blur'``.
          - Calculate difference between blended image and reflect image.
          - Calculate structure similarity (SSIM) between blended image and background image
            by calling :any:`skimage.metrics.structural_similarity`.
          - If the difference is relatively large enough, blended image is not very dark
            and SSIM is around ``(0.7, 0.85)``, current reflect image is added to candidates.
        * **Rank candidate triggers by conducting tentative attack with multiple triggers injected together.**

          - (Initialize, not repeated) Assign all candidate triggers with same sampling weights.
          - Sample certain amount (e.g., ``40%`` in original code) of clean data from training set in target class.
          - Randomly attach a candidate trigger on each clean input according to their sampling weights.
          - Use the infected data as poison dataset to retrain a pretrained model
            with :attr:`refool_epochs` and :attr:`refool_lr`.
          - Evaluate attack succ rate of each used trigger as their new sampling weights.
          - Set sampling weights of all unused triggers to the median of used ones.
          - Reset the model as pretrained state.
          - Repeat the ranking process for :attr:`rank_iter` times.
        * **Use the trigger with largest sampling weight for final attack**
          (with ``'dataset'`` train_mode).

    See Also:
        * paper: `Reflection Backdoor\: A Natural Backdoor Attack on Deep Neural Networks`_
        * code: https://github.com/DreamtaleCore/Refool

    Note:
        There are **differences** between our implementation and original codes.
        I've consulted first author to clarify that current implementation of TrojanZoo should work.

        * | Author's code allows repeat during generating candidate reflect images.
          | Our code has **NO** repeat.
        * | Author's code generates ``160`` (actually usually not reaching this number)
            candidate reflect images but requires ``200`` during attack, which causes more repeat.
          | Our code generate :attr:`candidate_num` (``100`` as default) unique candidates.
        * | Author's code uses a very large :attr:`refool_epochs` (``600``),
            which causes too much clean accuracy drop and is very slow.
          | Our code uses ``5`` as default.
        * | Author's code uses a very large :attr:`refool_sample_percent` (``0.4``),
            which causes too much clean accuracy drop.
          | Our code uses ``0.1`` as default.
        * | There should be a pretrained model that is reset at every ranking loop.
          | However, the paper and original code don't mention that.
          | The author tells me that they load pretrained model from ImageNet.
        * There is no attack code provided by original author after ranking candidate reflect images.

        There is also a **conflict** between codes and paper from original author.

        * | Paper claims to use top-:attr:`candidate_num` selection at every ranking loop in Algorithm 1.
          | Author's code uses random sampling according to ``W`` as sampling weights.
          | Our code follows **author's code**.

    Args:
        candidate_num (int): Number of candidate reflect images.
            Defaults to ``100``.
        rank_iter (int): Iteration to update sampling weights of candidate reflect images.
            Defaults to ``16``.
        refool_epochs (int): Retraining epochs during trigger ranking.
            Defaults to ``5``.
        refool_lr (float): Retraining learning rate during trigger ranking.
            Defaults to ``1e-3``.
        refool_sample_percent (float): Percentage of retraining samples
            by training set in target class during trigger ranking.
            Defaults to ``0.1``.
        voc_root (str): Path to Pascal VOC dataset.
            Defaults to ``'{data_dir}/image/voc'``.
        efficient (bool): Whether to only use a subset (20%) to evaluate ASR during trigger ranking.
            Defaults to ``False``.

    Attributes:
        reflect_imgs (torch.Tensor): Candidate reflect images with shape ``(candidate_num, C, H, W)``.
        train_mode (str): Training mode to inject backdoor. Forced to be 'dataset'.
            See detailed description in :class:`trojanvision.attacks.BadNet`.
        poison_set (torch.utils.data.Dataset): Poison dataset (no clean data).
            It is ``None`` at initialization because the best trigger keeps unknown.
        refool_sample_num (int): Number of retraining samples from training set
            in target class during trigger ranking.
            ``refool_sample_percent * len(target_set)``
        target_set (torch.utils.data.Dataset): Training set in target class.

    .. _Reflection Backdoor\: A Natural Backdoor Attack on Deep Neural Networks:
        https://arxiv.org/abs/2007.02343
    """  # noqa: E501
    name: str = 'refool'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--candidate_num', type=int,
                           help='number of candidate reflect images '
                           '(default: 100)')
        group.add_argument('--rank_iter', type=int,
                           help='iteration to update sampling weights of candidate reflect images '
                           '(default: 16)')
        group.add_argument('--refool_epochs', type=int,
                           help='retraining epochs during trigger ranking '
                           '(default: 5)')
        group.add_argument('--refool_lr', type=float,
                           help='retraining learning rate during trigger ranking '
                           '(default: 1e-3)')
        group.add_argument('--refool_sample_percent', type=int,
                           help='retraining samples by training set '
                           'in target class during trigger ranking '
                           '(default: 0.1)')
        group.add_argument('--voc_root', help='path to Pascal VOC dataset '
                           '(default: "{data_dir}/image/voc")')
        group.add_argument('--efficient', action='store_true',
                           help='whether to only use a subset (20%) '
                           'to evaluate ASR during trigger ranking')
        return group

    def __init__(self, candidate_num: int = 100, rank_iter: int = 16,
                 refool_epochs: int = 5, refool_lr: float = 1e-3,
                 refool_sample_percent: float = 0.1,
                 voc_root: str = None, efficient: bool = False,
                 **kwargs):
        super().__init__(**kwargs)
        self.param_list['refool'] = ['candidate_num', 'rank_iter', 'refool_epochs']
        self.candidate_num = candidate_num
        self.rank_iter = rank_iter
        self.refool_epochs = refool_epochs
        self.refool_lr = refool_lr
        self.refool_sample_percent = refool_sample_percent

        if voc_root is None:
            data_dir = os.path.dirname(os.path.dirname(self.dataset.folder_path))
            voc_root = os.path.join(data_dir, 'image', 'voc')
        self.voc_root = voc_root
        self.reflect_imgs = self._get_reflect_imgs()

        mark_shape = self.dataset.data_shape.copy()
        mark_shape[0] += 1
        self.mark.mark = torch.ones(mark_shape, device=self.mark.mark.device)
        self.mark.mark_height, self.mark.mark_width = self.dataset.data_shape[-2:]
        self.mark.mark_height_offset, self.mark.mark_width_offset = 0, 0
        self.mark.mark_random_init = False
        self.mark.mark_random_pos = False
        self.mark.mark_alpha = -1.0     # TODO: any manual alpha setting?

        self.refool_sample_num = int(self.refool_sample_percent * len(self.target_set))

        if efficient:
            valid_set = self.dataset.loader['valid'].dataset
            length = int(0.2 * len(valid_set))
            subset = torch.utils.data.Subset(valid_set, torch.randperm(len(valid_set))[:length])
            self.loader_valid = self.dataset.get_dataloader(mode='valid', dataset=subset)
        else:
            self.loader_valid = self.dataset.loader['valid']

    def add_mark(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        r"""Add watermark to input tensor by calling
        :meth:`trojanvision.attacks.BadNet.add_mark()`.

        If :attr:`mark_alpha` ``<0``, use mean of :attr:`x`
        and :attr:`self.mark.mark` as their weights.
        """
        mark_alpha = kwargs.get('mark_alpha', self.mark.mark_alpha)
        if mark_alpha < 0:
            x_weight: float = x.mean().item()
            mark_weight: float = self.mark.mark[:-1].mean().item()
            alpha = mark_weight / (x_weight + mark_weight)
            kwargs['mark_alpha'] = alpha
        return super().add_mark(x, **kwargs)

    def attack(self, epochs: int, optimizer: torch.optim.Optimizer, **kwargs):
        model_dict = copy.deepcopy(self.model.state_dict())
        W = torch.ones(len(self.reflect_imgs))
        refool_optimizer = torch.optim.SGD(optimizer.param_groups[0]['params'],
                                           lr=self.refool_lr, momentum=0.9,
                                           weight_decay=5e-4)
        # logger = MetricLogger(meter_length=35)
        # logger.create_meters(asr='{median:.3f} ({min:.3f}  {max:.3f})')
        # iterator = logger.log_every(range(self.rank_iter))
        for _iter in range(self.rank_iter):
            print('Select iteration: ', output_iter(_iter + 1, self.rank_iter))
            # prepare data
            idx = random.choices(range(len(W)), weights=W.tolist(), k=self.refool_sample_num)
            mark = torch.ones_like(self.mark.mark).expand(self.refool_sample_num, -1, -1, -1).clone()
            mark[:, :-1] = self.reflect_imgs[idx]
            clean_input, _ = sample_batch(self.target_set, self.refool_sample_num)
            trigger_input = self.add_mark(clean_input, mark=mark)
            dataset = TensorListDataset(trigger_input, [self.target_class] * len(trigger_input))
            loader = self.dataset.get_dataloader(mode='train', dataset=dataset)
            # train
            self.model._train(self.refool_epochs, optimizer=refool_optimizer,
                              loader_train=loader, validate_interval=0,
                              output_freq='epoch', indent=4)
            self.model._validate(indent=4)
            # test
            select_idx = list(set(idx))
            marks = self.reflect_imgs[select_idx]
            asr_result = self._get_asr_result(marks)
            # update W
            W[select_idx] = asr_result
            other_idx = list(set(range(len(W))) - set(idx))
            W[other_idx] = asr_result.median()

            # logger.reset().update_list(asr=asr_result)
            self.model.load_state_dict(model_dict)
        self.mark.mark[:-1] = self.reflect_imgs[W.argmax().item()]
        self.poison_set = self.get_poison_dataset(load_mark=False)
        return super().attack(epochs=epochs, optimizer=optimizer, **kwargs)

    def _get_asr_result(self, marks: torch.Tensor) -> torch.Tensor:
        r"""Get attack succ rate result for each mark in :attr:`marks`.

        Args:
            marks (torch.Tensor): Marks tensor with shape ``(N, C, H, W)``.

        Returns:
            torch.Tensor: Attack succ rate tensor with shape ``(N)``.
        """
        asr_list = []
        logger = MetricLogger(meter_length=35, indent=4)
        logger.create_meters(asr='{median:.3f} ({min:.3f}  {max:.3f})')
        for mark in logger.log_every(marks, header='mark', tqdm_header='mark'):
            self.mark.mark[:-1] = mark
            asr, _ = self.model._validate(get_data_fn=self.get_data, keep_org=False,
                                          poison_label=True, verbose=False,
                                          loader=self.loader_valid)
            # Original code considers an untargeted-like attack scenario.
            # org_acc, _ = self.model._validate(get_data_fn=self.get_data, keep_org=False,
            #                                   poison_label=False, verbose=False)
            # asr = 100 - org_acc
            logger.update(asr=asr)
            asr_list.append(asr)
        return torch.tensor(asr_list)

    def _get_reflect_imgs(self, force_regen: bool = False) -> torch.Tensor:
        r"""Get reflect images with shape ``(candidate_num, C, H, W)``.

        Will generate tar file containing reflect images
        if it doesn't exist or ``force_regen == True``.

        Args:
            force_regen (bool): Whether to force regenerating tar file.
                Defaults to ``False``.

        Returns:
            torch.Tensor: Reflect images with shape ``(N, C, H, W)``.
        """
        tar_path = os.path.join(self.voc_root, 'reflect.tar')
        if force_regen or not os.path.isfile(tar_path):
            gen_reflect_imgs(tar_path, self.voc_root, num_attack=self.candidate_num)
        tf = tarfile.open(tar_path, mode='r')
        transform = transforms.Compose([
            transforms.Resize([self.dataset.data_shape[-2:]]),
            transforms.PILToTensor(),
            transforms.ConvertImageDtype(torch.float)
        ])
        images = torch.stack([transform(Image.open(tf.extractfile(member), mode='r'))
                              for member in tf.getmembers()])
        if len(images) >= self.candidate_num:
            images = images[:self.candidate_num]
        elif not force_regen:
            return self._get_reflect_imgs(force_regen=True)
        else:
            raise RuntimeError('Can not generate enough images')
        tf.close()
        return images.to(device=env['device'])


def gen_reflect_imgs(tar_path: str, voc_root: str, num_attack: int = 160,
                     reflect_class: set[str] = {'cat'},
                     background_class: set[str] = {'person'}):
    r"""Generate a tar file containing reflect images.

    Args:
        tar_path (str): Tar file path to save.
        voc_root (str): VOC dataset root path.
        num_attack (int): Number of reflect images to generate.
        reflect_class (set[str]): Set of reflect classes.
        background_class (set[str]): Set of background classes.
    """
    print('get image paths')
    if not os.path.isdir(voc_root):
        os.makedirs(voc_root)
    datasets = [torchvision.datasets.VOCDetection(voc_root, year=year, image_set=image_set,
                                                  download=True) for year, image_set in sets]
    background_paths = _get_img_paths(datasets, positive_class=background_class, negative_class=reflect_class)
    reflect_paths = _get_img_paths(datasets, positive_class=reflect_class, negative_class=background_class)
    print()
    print('background: ', len(background_paths))
    print('reflect: ', len(reflect_paths))
    print()
    print('load images')
    reflect_imgs = [read_tensor(fp) for fp in reflect_paths]

    print('writing tar file: ', tar_path)
    tf = tarfile.open(tar_path, mode='w')
    logger = MetricLogger(meter_length=35)
    logger.create_meters(reflect_num=f'[ {{count:3d}} / {num_attack:3d} ]',
                         reflect_mean='{global_avg:.3f} ({min:.3f}  {max:.3f})',
                         diff_mean='{global_avg:.3f} ({min:.3f}  {max:.3f})',
                         blended_max='{global_avg:.3f} ({min:.3f}  {max:.3f})',
                         ssim='{global_avg:.3f} ({min:.3f}  {max:.3f})')
    candidates: set[int] = set()
    for fp in logger.log_every(background_paths):
        background_img = read_tensor(fp)
        for i, reflect_img in enumerate(reflect_imgs):
            if i in candidates:
                continue
            blended, background_layer, reflection_layer = blend_images(
                background_img, reflect_img, ghost_rate=0.39)
            reflect_mean: float = reflection_layer.mean().item()
            diff_mean: float = (blended - reflection_layer).mean().item()
            blended_max: float = blended.max().item()
            logger.update(reflect_mean=reflect_mean, diff_mean=diff_mean, blended_max=blended_max)
            if reflect_mean < 0.8 * diff_mean and blended_max > 0.1:
                ssim: float = skimage.metrics.structural_similarity(
                    blended.numpy(), background_layer.numpy(), channel_axis=0)
                logger.update(ssim=ssim)
                if 0.7 < ssim < 0.85:
                    logger.update(reflect_num=1)
                    candidates.add(i)
                    filename = os.path.basename(reflect_paths[i])
                    bytes_io = io.BytesIO()
                    format = os.path.splitext(filename)[1][1:].lower().replace('jpg', 'jpeg')
                    torchvision.utils.save_image(reflection_layer, bytes_io, format=format)
                    bytes_data = bytes_io.getvalue()
                    tarinfo = tarfile.TarInfo(name=filename)
                    tarinfo.size = len(bytes_data)
                    tf.addfile(tarinfo, io.BytesIO(bytes_data))
                    break
        if len(candidates) == num_attack:
            break
    else:
        raise RuntimeError('Can not generate enough images')
    tf.close()


def _get_img_paths(datasets: list[torchvision.datasets.VOCDetection],
                   positive_class: set[str], negative_class: set[str]
                   ) -> list[str]:
    r"""Get image paths that contain at least 1 object in :attr:`positive_class`
    and no object in :attr:`negative_class`.

    Args:
        datasets: (list[torchvision.datasets.VOCDetection]):
            list of different VOC datasets.
        positive_class (set[str]): Selected image should contain
            at least 1 object in :attr:`positive_class`.
        negative_class (set[str]): Selected image should **NOT** contain
            any object in :attr:`negative_class`.

    Returns:
        list[str]: List of selected image paths.
    """
    image_paths: list[str] = []
    for dataset in datasets:
        for index in range(len(dataset)):
            target = dataset.parse_voc_xml(ET_parse(dataset.annotations[index]).getroot())
            label_names: set[str] = {obj['name'] for obj in target['annotation']['object']}
            if len(positive_class & label_names) != 0 and len(negative_class & label_names) == 0:
                image_paths.append(dataset.images[index])
    return image_paths


def blend_images(background_img: torch.Tensor, reflect_img: torch.Tensor,
                 max_image_size: int = 560, ghost_rate: float = 0.49, alpha_bg: float = None,
                 offset: tuple[int, int] = (0, 0), sigma: float = None, ghost_alpha: float = None
                 ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Blend background layer and reflection layer together (including blurred & ghosted reflection layer).

    :attr:`background_img` is resized using :attr:`max_image_size`
    and :attr:`reflect_img` is resized to the same shape.

    Note:
        This blend method is only used to generate reflect images.
        To add watermark on images, please call :meth:`add_mark()`.

    Args:
        background_img (torch.Tensor): Background image tensor with shape ``([N], C, H, W)``.
        reflect_img (torch.Tensor): Reflect image tensor with shape ``([N], C, H', W')``.
        max_image_size (int): Max image size (the longer edge of height/width).
            :attr:`background_img` will be resized while keeping aspect ratio.
            Defaults to ``560``.
        ghost_rate (float): Probability to generate the blended image with ghost effect.
            Defaults to ``0.49``.
        alpha_bg (float): Weight of background image during blending.
            Defaults to ``1 - random.uniform(0.05, 0.45)``。
        offset (tuple[int, int]): Offset of height and width used in ghost effect.
            Defaults to ``(random.randint(3, 8), random.randint(3, 8))``。
        sigma (float): Gaussian kernel standard deviation.
            Defaults to ``random.uniform(1, 5)``.
        ghost_alpha (float): Weight of the first ghost image used in ghost effect.
            Defaults to ``abs(round(random.random()) - random.uniform(0.15, 0.5))``.

    Returns:
        (torch.Tensor, torch.Tensor, torch.Tensor):
            ``blended, background_layer, reflection_layer``
            with shape ``([N], C, H, W)``.
    """
    if alpha_bg is None:
        alpha_bg = 1. - random.uniform(0.05, 0.45)
    h, w = background_img.shape[-2:]
    aspect_ratio = w / h
    h, w = (max_image_size, int(round(max_image_size * aspect_ratio))) if h > w \
        else (int(round(max_image_size / aspect_ratio)), max_image_size)
    # Original code uses cv2 INTER_CUBIC, which is slightly different from BICUBIC
    background_img = F.resize(background_img, size=(h, w), interpolation=InterpolationMode.BICUBIC).clamp(0, 1)
    reflect_img = F.resize(reflect_img, size=(h, w), interpolation=InterpolationMode.BICUBIC).clamp(0, 1)
    background_img.pow_(2.2)
    reflect_img.pow_(2.2)

    background_mask = alpha_bg * background_img
    if random.random() < ghost_rate:
        # generate the blended image with ghost effect
        if ghost_alpha is None:
            ghost_alpha = abs(round(random.random()) - random.uniform(0.15, 0.5))
        if offset[0] == 0 and offset[1] == 0:
            offset = (random.randint(3, 8), random.randint(3, 8))
        reflect_1 = F.pad(background_img, [0, 0, offset[0], offset[1]])  # pad on right/bottom
        reflect_2 = F.pad(background_img, [offset[0], offset[1], 0, 0])  # pad on left/top
        reflect_ghost = ghost_alpha * reflect_1 + (1 - ghost_alpha) * reflect_2
        reflect_ghost = reflect_ghost[..., offset[0]: -offset[0], offset[1]: -offset[1]]
        reflect_ghost = F.resize(reflect_ghost, size=[h, w],
                                 interpolation=InterpolationMode.BICUBIC
                                 ).clamp(0, 1)  # no cubic mode in original code

        reflect_mask = (1 - alpha_bg) * reflect_ghost
        reflection_layer = reflect_mask.pow(1 / 2.2)
    else:
        # generate the blended image with focal blur
        if sigma is None:
            sigma = random.uniform(1, 5)
        kernel_size = int(2 * math.ceil(2 * sigma) + 1)
        reflect_blur = F.gaussian_blur(reflect_img, kernel_size, sigma)
        blend = reflect_blur + background_img

        # get the reflection layers' proper range
        att = 1.08 + random.random() / 10.0
        mask = blend > 1
        mean = torch.tensor([blend[i, mask[i]].mean().nan_to_num(1.0).item()
                            for i in range(blend.size(0))]).view(-1, 1, 1)    # (C, 1, 1)
        reflect_blur = (reflect_blur - att * (mean - 1)).clamp(0, 1)

        def gen_kernel(kern_len: int = 100, nsig: int = 1) -> torch.Tensor:
            r"""Returns a 2D Gaussian kernel tensor."""
            interval = (2 * nsig + 1.) / kern_len
            x = torch.linspace(-nsig - interval / 2., nsig + interval / 2., kern_len + 1)
            # get normal distribution
            kern1d = norm.cdf(x).diff()
            kernel_raw = kern1d.outer(kern1d).sqrt()
            kernel = kernel_raw / kernel_raw.sum()  # TODO: is it auxiliary for positive numbers?
            kernel = kernel / kernel.max()
            return kernel
        h, w = reflect_blur.shape[-2:]
        new_h = random.randint(0, max_image_size - h - 10) if h < max_image_size - 10 else 0
        new_w = random.randint(0, max_image_size - w - 10) if w < max_image_size - 10 else 0
        g_mask = gen_kernel(max_image_size, 3).repeat(3, 1, 1)  # TODO: try to avoid hard encode 3 as channel
        alpha_r = (1 - alpha_bg / 2) * g_mask[..., new_h: new_h + h, new_w: new_w + w]

        reflect_mask = alpha_r * reflect_blur
        reflection_layer = (min(1., 4 * (1 - alpha_bg)) * reflect_mask).pow(1 / 2.2)

    blended = (reflect_mask + background_mask).pow(1 / 2.2)
    background_layer = background_mask.pow(1 / 2.2)
    return blended, background_layer, reflection_layer



# Used for python interactive window

import trojanvision

env = trojanvision.environ.create(verbose=1, color=True, tqdm=True)
dataset = trojanvision.datasets.create()
model = trojanvision.models.create(dataset=dataset)
trainer = trojanvision.trainer.create(dataset=dataset, model=model)
mark = trojanvision.marks.create(dataset=dataset)
attack = trojanvision.attacks.create('trojannn', dataset=dataset, model=model, mark=mark)



#!/usr/bin/env python3

from ...abstract import BackdoorDefense

import torch
import torchvision.transforms.functional as F
import argparse


class Recompress(BackdoorDefense):
    name: str = 'recompress'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--resize_ratio', type=float, help='Image Resize Ratio for Recompress, defaults to 0.95.')
        return group

    def __init__(self, resize_ratio: float = 0.95, **kwargs):
        super().__init__(**kwargs)
        self.param_list['recompress'] = ['resize_ratio']
        self.resize_ratio = resize_ratio

    def detect(self, **kwargs):
        super().detect(**kwargs)
        self.validate_fn()

    def validate_fn(self, **kwargs) -> tuple[float, float]:
        clean_acc, _ = self.model._validate(print_prefix='Validate Clean',
                                            get_data_fn=self.get_data, org=True, **kwargs)
        asr, _ = self.model._validate(print_prefix='Validate ASR',
                                      get_data_fn=self.get_data, keep_org=False, **kwargs)
        # self.model._validate(print_prefix='Validate Trigger Org',
        #                      get_data_fn=self.get_data, keep_org=False, poison_label=False, **kwargs)
        # print(f'Validate Confidence : {self.attack.validate_confidence():.3f}')
        return asr, clean_acc

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor],
                 **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.attack.get_data(data=data, **kwargs)
        h, w = _input.shape[-2], _input.shape[-1]
        _input = F.resize(_input, (int(h * self.resize_ratio), int(w * self.resize_ratio)))
        _input = F.resize(_input, (h, w))
        return _input, _label



#!/usr/bin/env python3

from ...abstract import InputFiltering
from .neo import Neo
from .strip import Strip

__all__ = ['Neo', 'Strip']

class_dict: dict[str, type[InputFiltering]] = {
    'neo': Neo,
    'strip': Strip,
}



#!/usr/bin/env python3

from trojanzoo.utils.fim import KFAC, EKFAC
from trojanzoo.utils.logger import MetricLogger
from trojanzoo.utils.memory import empty_cache
from trojanzoo.utils.model import accuracy, activate_params
from trojanzoo.utils.output import ansi, get_ansi_len, output_iter, prints
from trojanzoo.environ import env

import torch
import torch.nn as nn
import torch.nn.functional as F

from tqdm import tqdm

from trojanzoo.utils.model import ExponentialMovingAverage
from collections.abc import Callable
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
import torch.utils.data


def train(module: nn.Module, num_classes: int,
          epochs: int, optimizer: Optimizer, lr_scheduler: _LRScheduler = None,
          lr_warmup_epochs: int = 0,
          model_ema: ExponentialMovingAverage = None,
          model_ema_steps: int = 32,
          grad_clip: float = None, pre_conditioner: None | KFAC | EKFAC = None,
          print_prefix: str = 'Train', start_epoch: int = 0, resume: int = 0,
          validate_interval: int = 10, save: bool = False, amp: bool = False,
          loader_train: torch.utils.data.DataLoader = None,
          loader_valid: torch.utils.data.DataLoader = None,
          epoch_fn: Callable[..., None] = None,
          get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
          forward_fn: Callable[..., torch.Tensor] = None,
          loss_fn: Callable[..., torch.Tensor] = None,
          after_loss_fn: Callable[..., None] = None,
          validate_fn: Callable[..., tuple[float, float]] = None,
          save_fn: Callable[..., None] = None, file_path: str = None,
          folder_path: str = None, suffix: str = None,
          writer=None, main_tag: str = 'train', tag: str = '',
          accuracy_fn: Callable[..., list[float]] = None,
          verbose: bool = True, output_freq: str = 'iter', indent: int = 0,
          change_train_eval: bool = True, lr_scheduler_freq: str = 'epoch',
          backward_and_step: bool = True,
          **kwargs):
    r"""Train the model"""
    if epochs <= 0:
        return
    get_data_fn = get_data_fn or (lambda x: x)
    forward_fn = forward_fn or module.__call__
    loss_fn = loss_fn or (lambda _input, _label, _output=None: F.cross_entropy(_output or forward_fn(_input), _label))
    validate_fn = validate_fn or validate
    accuracy_fn = accuracy_fn or accuracy

    scaler: torch.cuda.amp.GradScaler = None
    if not env['num_gpus']:
        amp = False
    if amp:
        scaler = torch.cuda.amp.GradScaler()
    best_validate_result = (0.0, float('inf'))
    if validate_interval != 0:
        best_validate_result = validate_fn(loader=loader_valid, get_data_fn=get_data_fn,
                                           forward_fn=forward_fn, loss_fn=loss_fn,
                                           writer=writer, tag=tag, _epoch=start_epoch,
                                           verbose=verbose, indent=indent, **kwargs)
        best_acc = best_validate_result[0]

    params: list[nn.Parameter] = []
    for param_group in optimizer.param_groups:
        params.extend(param_group['params'])
    len_loader_train = len(loader_train)
    total_iter = (epochs - resume) * len_loader_train

    logger = MetricLogger()
    logger.create_meters(loss=None, top1=None, top5=None)

    if resume and lr_scheduler:
        for _ in range(resume):
            lr_scheduler.step()
    iterator = range(resume, epochs)
    if verbose and output_freq == 'epoch':
        header: str = '{blue_light}{0}: {reset}'.format(print_prefix, **ansi)
        header = header.ljust(max(len(header), 30) + get_ansi_len(header))
        iterator = logger.log_every(range(resume, epochs),
                                    header=print_prefix,
                                    tqdm_header='Epoch',
                                    indent=indent)
    for _epoch in iterator:
        _epoch += 1
        logger.reset()
        if callable(epoch_fn):
            activate_params(module, [])
            epoch_fn(optimizer=optimizer, lr_scheduler=lr_scheduler,
                     _epoch=_epoch, epochs=epochs, start_epoch=start_epoch)
        loader_epoch = loader_train
        if verbose and output_freq == 'iter':
            header: str = '{blue_light}{0}: {1}{reset}'.format(
                'Epoch', output_iter(_epoch, epochs), **ansi)
            header = header.ljust(max(len('Epoch'), 30) + get_ansi_len(header))
            loader_epoch = logger.log_every(loader_train, header=header,
                                            tqdm_header='Batch',
                                            indent=indent)
        if change_train_eval:
            module.train()
        activate_params(module, params)
        for i, data in enumerate(loader_epoch):
            _iter = _epoch * len_loader_train + i
            # data_time.update(time.perf_counter() - end)
            _input, _label = get_data_fn(data, mode='train')
            if pre_conditioner is not None and not amp:
                pre_conditioner.track.enable()
            _output = forward_fn(_input, amp=amp, parallel=True)
            loss = loss_fn(_input, _label, _output=_output, amp=amp)
            if backward_and_step:
                optimizer.zero_grad()
                if amp:
                    scaler.scale(loss).backward()
                    if callable(after_loss_fn) or grad_clip is not None:
                        scaler.unscale_(optimizer)
                    if callable(after_loss_fn):
                        after_loss_fn(_input=_input, _label=_label,
                                      _output=_output,
                                      loss=loss, optimizer=optimizer,
                                      loss_fn=loss_fn,
                                      amp=amp, scaler=scaler,
                                      _iter=_iter, total_iter=total_iter)
                    if grad_clip is not None:
                        nn.utils.clip_grad_norm_(params, grad_clip)
                    scaler.step(optimizer)
                    scaler.update()
                else:
                    loss.backward()
                    if callable(after_loss_fn):
                        after_loss_fn(_input=_input, _label=_label,
                                      _output=_output,
                                      loss=loss, optimizer=optimizer,
                                      loss_fn=loss_fn,
                                      amp=amp, scaler=scaler,
                                      _iter=_iter, total_iter=total_iter)
                        # start_epoch=start_epoch, _epoch=_epoch, epochs=epochs)
                    if pre_conditioner is not None:
                        pre_conditioner.track.disable()
                        pre_conditioner.step()
                    if grad_clip is not None:
                        nn.utils.clip_grad_norm_(params, grad_clip)
                    optimizer.step()

            if model_ema and i % model_ema_steps == 0:
                model_ema.update_parameters(module)
                if _epoch <= lr_warmup_epochs:
                    # Reset ema buffer to keep copying weights
                    # during warmup period
                    model_ema.n_averaged.fill_(0)

            if lr_scheduler and lr_scheduler_freq == 'iter':
                lr_scheduler.step()
            acc1, acc5 = accuracy_fn(
                _output, _label, num_classes=num_classes, topk=(1, 5))
            batch_size = int(_label.size(0))
            logger.update(n=batch_size, loss=float(loss), top1=acc1, top5=acc5)
            empty_cache()
        optimizer.zero_grad()
        if lr_scheduler and lr_scheduler_freq == 'epoch':
            lr_scheduler.step()
        if change_train_eval:
            module.eval()
        activate_params(module, [])
        loss, acc = (logger.meters['loss'].global_avg,
                     logger.meters['top1'].global_avg)
        if writer is not None:
            from torch.utils.tensorboard import SummaryWriter
            assert isinstance(writer, SummaryWriter)
            writer.add_scalars(main_tag='Loss/' + main_tag,
                               tag_scalar_dict={tag: loss},
                               global_step=_epoch + start_epoch)
            writer.add_scalars(main_tag='Acc/' + main_tag,
                               tag_scalar_dict={tag: acc},
                               global_step=_epoch + start_epoch)
        if validate_interval != 0 and (_epoch % validate_interval == 0 or _epoch == epochs):
            validate_result = validate_fn(module=module,
                                          num_classes=num_classes,
                                          loader=loader_valid,
                                          get_data_fn=get_data_fn,
                                          forward_fn=forward_fn,
                                          loss_fn=loss_fn,
                                          writer=writer, tag=tag,
                                          _epoch=_epoch + start_epoch,
                                          verbose=verbose, indent=indent,
                                          **kwargs)
            cur_acc = validate_result[0]
            if cur_acc >= best_acc:
                best_validate_result = validate_result
                if verbose:
                    prints('{purple}best result update!{reset}'.format(
                        **ansi), indent=indent)
                    prints(f'Current Acc: {cur_acc:.3f}    '
                           f'Previous Best Acc: {best_acc:.3f}',
                           indent=indent)
                best_acc = cur_acc
                if save:
                    save_fn(file_path=file_path, folder_path=folder_path,
                            suffix=suffix, verbose=verbose)
            if verbose:
                prints('-' * 50, indent=indent)
    module.zero_grad()
    return best_validate_result


def validate(module: nn.Module, num_classes: int,
             loader: torch.utils.data.DataLoader,
             print_prefix: str = 'Validate', indent: int = 0,
             verbose: bool = True,
             get_data_fn: Callable[
                 ..., tuple[torch.Tensor, torch.Tensor]] = None,
             forward_fn: Callable[..., torch.Tensor] = None,
             loss_fn: Callable[..., torch.Tensor] = None,
             writer=None, main_tag: str = 'valid',
             tag: str = '', _epoch: int = None,
             accuracy_fn: Callable[..., list[float]] = None,
             **kwargs) -> tuple[float, float]:
    r"""Evaluate the model.

    Returns:
        (float, float): Accuracy and loss.
    """
    module.eval()
    get_data_fn = get_data_fn or (lambda x: x)
    forward_fn = forward_fn or module.__call__
    loss_fn = loss_fn or nn.CrossEntropyLoss()
    accuracy_fn = accuracy_fn or accuracy
    logger = MetricLogger()
    logger.create_meters(loss=None, top1=None, top5=None)
    loader_epoch = loader
    if verbose:
        header: str = '{yellow}{0}{reset}'.format(print_prefix, **ansi)
        header = header.ljust(max(len(print_prefix), 30) + get_ansi_len(header))
        loader_epoch = logger.log_every(loader, header=header,
                                        tqdm_header='Batch',
                                        indent=indent)
    for data in loader_epoch:
        _input, _label = get_data_fn(data, mode='valid', **kwargs)
        with torch.no_grad():
            _output = forward_fn(_input)
            loss = float(loss_fn(_input, _label, _output=_output, **kwargs))
            acc1, acc5 = accuracy_fn(
                _output, _label, num_classes=num_classes, topk=(1, 5))
            batch_size = int(_label.size(0))
            logger.update(n=batch_size, loss=float(loss), top1=acc1, top5=acc5)
    acc, loss = (logger.meters['top1'].global_avg,
                 logger.meters['loss'].global_avg)
    if writer is not None and _epoch is not None and main_tag:
        from torch.utils.tensorboard import SummaryWriter
        assert isinstance(writer, SummaryWriter)
        writer.add_scalars(main_tag='Acc/' + main_tag,
                           tag_scalar_dict={tag: acc}, global_step=_epoch)
        writer.add_scalars(main_tag='Loss/' + main_tag,
                           tag_scalar_dict={tag: loss}, global_step=_epoch)
    return acc, loss


@torch.no_grad()
def compare(module1: nn.Module, module2: nn.Module,
            loader: torch.utils.data.DataLoader,
            print_prefix='Validate', indent=0, verbose=True,
            get_data_fn: Callable[...,
                                  tuple[torch.Tensor, torch.Tensor]] = None,
            criterion: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
            **kwargs) -> float:
    module1.eval()
    module2.eval()
    get_data_fn = get_data_fn if get_data_fn is not None else lambda x: x

    logger = MetricLogger()
    logger.create_meters(loss=None)
    loader_epoch = loader
    if verbose:
        header: str = '{yellow}{0}{reset}'.format(print_prefix, **ansi)
        header = header.ljust(
            max(len(print_prefix), 30) + get_ansi_len(header))
        if env['tqdm']:
            loader_epoch = tqdm(loader_epoch, leave=False)
        loader_epoch = logger.log_every(
            loader_epoch, header=header, indent=indent)
    for data in loader_epoch:
        _input, _label = get_data_fn(data, **kwargs)
        _output1: torch.Tensor = module1(_input)
        _output2: torch.Tensor = module2(_input)
        loss = criterion(_output1, _output2.softmax(1)).item()
        batch_size = int(_label.size(0))
        logger.update(n=batch_size, loss=loss)
    return logger.meters['loss'].global_avg



#!/usr/bin/env python3

from trojanvision.environ import env
from trojanzoo.attacks import Attack
from trojanzoo.optim import PGD as PGDoptimizer
from trojanzoo.utils.output import prints, ansi
from trojanzoo.utils.logger import SmoothedValue

from trojanvision.datasets import ImageSet
import torch
import argparse
from collections.abc import Callable
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from trojanvision.models import ImageModel


class PGD(Attack, PGDoptimizer):
    r"""PGD Adversarial Attack.

    Args:
        pgd_alpha (float): learning rate :math:`\alpha`. Default: :math:`\frac{3}{255}`.
        pgd_eps (float): the perturbation threshold :math:`\epsilon` in input space. Default: :math:`\frac{8}{255}`.
    """

    name: str = 'pgd'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--pgd_alpha', type=float, help='PGD learning rate per step, defaults to 2.0/255')
        group.add_argument('--pgd_eps', type=float, help='Projection norm constraint, defaults to 8.0/255')
        group.add_argument('--iteration', type=int, help='Attack Iteration, defaults to 7')
        group.add_argument('--stop_threshold', type=float, help='early stop confidence, defaults to 0.99 (defined in config)')
        group.add_argument('--target_class', type=int, help='Do not set it if using target_idx')
        group.add_argument('--target_idx', type=int,
                           help='Target label order in original classification, defaults to -1 '
                           '(0 for untargeted attack, 1 for most possible class, -1 for most unpossible class)')
        group.add_argument('--test_num', type=int, help='total number of test examples for PGD, defaults to 1000.')
        group.add_argument('--num_restart', type=int,
                           help='number of random init for PGD, defaults to 0 (without random initialization).')
        group.add_argument('--require_class', action='store_true',
                           help='Require class to be desired in early_stop_check.')

        group.add_argument('--grad_method', choices=['white', 'nes', 'sgd', 'hess', 'zoo'],
                           help='gradient estimation method, defaults to "white"')
        group.add_argument('--query_num', type=int,
                           help='query numbers for black box gradient estimation, defaults to 100.')
        group.add_argument('--sigma', type=float,
                           help='gaussian sampling std for black box gradient estimation, defaults to 1e-3')
        return group

    def __init__(self, target_class: int = None, target_idx: int = -1, test_num: int = 1000, num_restart: int = 0,
                 require_class: bool = False,
                 dataset: ImageSet = None, model: 'ImageModel' = None,
                 pgd_alpha: float | torch.Tensor = 2.0 / 255,
                 pgd_eps: float | torch.Tensor = 8.0 / 255,
                 clip_min: float | torch.Tensor = 0.0,
                 clip_max: float | torch.Tensor = 1.0,
                 **kwargs):
        self.target_class = target_class
        self.target_idx = target_idx
        self.test_num = test_num
        self.num_restart = num_restart
        self.require_class = require_class
        if num_restart > 0:
            kwargs.update(random_init=True)

        if model.norm_par is None and isinstance(dataset, ImageSet):
            if dataset.normalize and dataset.norm_par is not None:
                mean = torch.tensor(dataset.norm_par['mean'],
                                    device=env['device']).view(-1, 1, 1)
                std = torch.tensor(dataset.norm_par['std'],
                                   device=env['device']).view(-1, 1, 1)
                clip_min = (clip_min - mean) / std
                clip_max = (clip_max - mean) / std
                pgd_alpha = pgd_alpha / std
                pgd_eps = pgd_eps / std
        super().__init__(dataset=dataset, model=model,
                         pgd_alpha=pgd_alpha, pgd_eps=pgd_eps,
                         clip_min=clip_min, clip_max=clip_max, **kwargs)
        self.param_list['pgd_attack'] = ['target_class', 'target_idx', 'test_num', 'num_restart', 'require_class']
        self.dataset: ImageSet
        self.model: 'ImageModel'

    def attack(self, verbose: int = 1, **kwargs) -> tuple[float, float]:
        validset = self.dataset.get_dataset('valid')
        testset, _ = self.dataset.split_dataset(validset, percent=0.3)
        loader = self.dataset.get_dataloader(mode='valid', dataset=testset,
                                             shuffle=True)
        fmt_str = '{global_avg:7.3f} ({min:7.3f}  {max:7.3f})'
        total_adv_target_conf = SmoothedValue(fmt=fmt_str)
        total_org_target_conf = SmoothedValue(fmt=fmt_str)
        succ_adv_target_conf = SmoothedValue(fmt=fmt_str)

        total_adv_org_conf = SmoothedValue(fmt=fmt_str)
        total_org_org_conf = SmoothedValue(fmt=fmt_str)
        succ_adv_org_conf = SmoothedValue(fmt=fmt_str)

        total_iter_list = SmoothedValue(fmt=fmt_str)
        succ_iter_list = SmoothedValue(fmt=fmt_str)

        succ_idx_list: list[int] = []
        for data in loader:
            rest_length = self.test_num - total_adv_target_conf.count
            if rest_length <= 0:
                break
            _input, _label = self.model.remove_misclassify(data)
            if len(_label) == 0:
                continue

            if len(_label) > rest_length:
                _input = _input[:rest_length]
                _label = _label[:rest_length]
            target = self.generate_target(_input, idx=self.target_idx) if self.target_class is None \
                else self.target_class * torch.ones_like(_label)
            adv_input = _input.clone().detach()
            iter_list = -torch.ones(len(_label), dtype=torch.long)
            current_idx = torch.arange(len(iter_list))
            for _ in range(max(self.num_restart, 1)):
                temp_adv_input, temp_iter_list = self.optimize(_input[current_idx],
                                                               target=target[current_idx], **kwargs)
                adv_input[current_idx] = temp_adv_input
                iter_list[current_idx] = temp_iter_list
                fail_idx = iter_list == -1
                if (~fail_idx).all():
                    break
                current_idx = current_idx[fail_idx]
            for i, _iter in enumerate(iter_list):
                if _iter != -1:
                    succ_idx_list.append(total_iter_list.count + i)
            adv_target_conf = self.model.get_target_prob(adv_input, target)
            adv_org_conf = self.model.get_target_prob(adv_input, _label)
            org_target_conf = self.model.get_target_prob(_input, target)
            org_org_conf = self.model.get_target_prob(_input, _label)

            total_adv_target_conf.update_list(adv_target_conf.detach().cpu().tolist())
            total_adv_org_conf.update_list(adv_org_conf.detach().cpu().tolist())
            succ_adv_target_conf.update_list(adv_target_conf[iter_list != -1].detach().cpu().tolist())
            succ_adv_org_conf.update_list(adv_org_conf[iter_list != -1].detach().cpu().tolist())
            total_org_target_conf.update_list(org_target_conf.detach().cpu().tolist())
            total_org_org_conf.update_list(org_org_conf.detach().cpu().tolist())

            total_iter_list.update_list(torch.where(iter_list != -1, iter_list, 2 *
                                        self.iteration * torch.ones_like(iter_list)).tolist())
            succ_iter_list.update_list(iter_list[iter_list != -1].tolist())
            if verbose >= 3:
                prints(f'{ansi["green"]}{succ_iter_list.count} / {total_iter_list.count}{ansi["reset"]}')
            if verbose >= 4:
                prints(f'{total_iter_list=:}', indent=4)
                prints(f'{succ_iter_list=:}', indent=4)
                prints()
                prints('-------------------------------------------------', indent=4)
                prints(f'{ansi["yellow"]}Target Class:{ansi["reset"]}', indent=4)
                prints(f'{total_adv_target_conf=:}', indent=8)
                prints(f'{total_org_target_conf=:}', indent=8)
                prints(f'{succ_adv_target_conf=:}', indent=8)
                prints()
                prints('-------------------------------------------------', indent=4)
                prints(f'{ansi["yellow"]}Original Class:{ansi["reset"]}', indent=4)
                prints(f'{total_adv_org_conf=:}', indent=8)
                prints(f'{total_org_org_conf=:}', indent=8)
                prints(f'{succ_adv_org_conf=:}', indent=8)
        if verbose:
            prints(f'{ansi["green"]}{succ_iter_list.count} / {total_iter_list.count}{ansi["reset"]}')
            prints(succ_idx_list)
        if verbose >= 2:
            prints(f'{total_iter_list=:}', indent=4)
            prints(f'{succ_iter_list=:}', indent=4)
            prints()
            prints('-------------------------------------------------', indent=4)
            prints(f'{ansi["yellow"]}Target Class:{ansi["reset"]}', indent=4)
            prints(f'{total_adv_target_conf=:}', indent=8)
            prints(f'{total_org_target_conf=:}', indent=8)
            prints(f'{succ_adv_target_conf=:}', indent=8)
            prints()
            prints('-------------------------------------------------', indent=4)
            prints(f'{ansi["yellow"]}Original Class:{ansi["reset"]}', indent=4)
            prints(f'{total_adv_org_conf=:}', indent=8)
            prints(f'{total_org_org_conf=:}', indent=8)
            prints(f'{succ_adv_org_conf=:}', indent=8)
        return float(succ_iter_list.count) / total_iter_list.count, total_iter_list.global_avg

    def optimize(self, _input: torch.Tensor, *args,
                 target: None | int | torch.Tensor = None, target_idx: int = None,
                 loss_fn: Callable[..., torch.Tensor] = None,
                 require_class: bool = None,
                 loss_kwargs: dict[str, torch.Tensor] = {},
                 **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        if len(_input) == 0:
            return _input, None
        target_idx = self.target_idx if target_idx is None else target_idx
        match target:
            case None:
                target = self.generate_target(_input, idx=self.target_idx) if self.target_class is None \
                    else self.target_class * torch.ones(len(_input), dtype=torch.long, device=_input.device)
            case int():
                target = target * torch.ones(len(_input), dtype=torch.long, device=_input.device)
        if loss_fn is None and self.loss_fn is None:
            untarget_condition = self.target_class is None and self.target_idx == 0

            def _loss_fn(_input: torch.Tensor, target: torch.Tensor, reduction: str = 'mean', **kwargs):
                loss = self.model.loss(_input, target, reduction=reduction)
                return -loss if untarget_condition else loss
            loss_fn = _loss_fn
        loss_kwargs.update(target=target)
        return super().optimize(_input, *args, target=target,
                                loss_fn=loss_fn, require_class=require_class,
                                loss_kwargs=loss_kwargs,
                                **kwargs)

    @torch.no_grad()
    def early_stop_check(self, current_idx: torch.Tensor,
                         adv_input: torch.Tensor, target: torch.Tensor, *args,
                         stop_threshold: float = None, require_class: bool = None,
                         **kwargs) -> torch.Tensor:
        stop_threshold = stop_threshold if stop_threshold is not None else self.stop_threshold
        require_class = require_class if require_class is not None else self.require_class
        if self.stop_threshold is None:
            return torch.zeros(len(current_idx), dtype=torch.bool)
        _confidence = self.model.get_target_prob(adv_input[current_idx], target[current_idx])
        untarget_condition = self.target_class is None and self.target_idx == 0
        result = _confidence > stop_threshold
        if untarget_condition:
            result = ~result
        if require_class:
            _class = self.model.get_class(adv_input[current_idx])
            class_result = _class == target[current_idx]
            if untarget_condition:
                class_result = ~class_result
            result = result.bitwise_and(class_result)
        return result.detach().cpu()

    @torch.no_grad()
    def output_info(self, org_input: torch.Tensor, noise: torch.Tensor, target: torch.Tensor,
                    loss_fn: Callable[[torch.Tensor], torch.Tensor] = None, **kwargs):
        super().output_info(org_input=org_input, noise=noise, loss_fn=loss_fn, **kwargs)
        # prints('Original class     : ', _label, indent=self.indent)
        # prints('Original confidence: ', _confidence, indent=self.indent)
        _confidence = self.model.get_target_prob(org_input + noise, target)
        prints('Target   class     : ', target.detach().cpu().tolist(), indent=self.indent)
        prints('Target   confidence: ', _confidence.detach().cpu().tolist(), indent=self.indent)



#!/usr/bin/env python3

from ...abstract import BackdoorDefense
from .abs import ABS
from .deep_inspect import DeepInspect
from .neural_cleanse import NeuralCleanse
from .neuron_inspect import NeuronInspect
from .tabor import Tabor

__all__ = ['ABS', 'DeepInspect', 'NeuralCleanse', 'NeuronInspect', 'Tabor']

class_dict: dict[str, type[BackdoorDefense]] = {
    'abs': ABS,
    'deep_inspect': DeepInspect,
    'neural_cleanse': NeuralCleanse,
    'neuron_inspect': NeuronInspect,
    'tabor': Tabor,
}



#!/usr/bin/env python3

import torch
import torch.nn as nn
import torchvision.transforms as transforms
from torchvision.transforms import functional as F
from torch.nn.functional import one_hot

import math


__all__ = ['RandomMixup', 'RandomCutmix', 'Cutout',
           'get_transform_bit',
           'get_transform_imagenet',
           'get_transform_cifar']


class RandomMixup(nn.Module):
    """Randomly apply Mixup to the provided batch and targets.
    The class implements the data augmentations as described in the paper
    `"mixup: Beyond Empirical Risk Minimization" <https://arxiv.org/abs/1710.09412>`_.
    Args:
        num_classes (int): number of classes used for one-hot encoding.
        p (float): probability of the batch being transformed. Default value is 0.5.
        alpha (float): hyperparameter of the Beta distribution used for mixup.
            Default value is 1.0.
        inplace (bool): boolean to make this transform inplace. Default set to False.
    """  # noqa: E501

    def __init__(self, num_classes: int, p: float = 0.5,
                 alpha: float = 1.0, inplace: bool = False) -> None:
        super().__init__()
        assert num_classes > 0, ("Please provide a valid positive value"
                                 " for the num_classes.")
        assert alpha > 0, "Alpha param can't be zero."

        self.num_classes = num_classes
        self.p = p
        self.alpha = alpha
        self.inplace = inplace

    def forward(self, batch: torch.Tensor,
                target: torch.Tensor
                ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Args:
            batch (torch.Tensor): Float tensor of size (B, C, H, W)
            target (torch.Tensor): Integer tensor of size (B, )
        Returns:
            torch.Tensor: Randomly transformed batch.
        """
        if batch.ndim != 4:
            raise ValueError(f"Batch ndim should be 4. Got {batch.ndim}")
        if target.ndim != 1:
            raise ValueError(f"Target ndim should be 1. Got {target.ndim}")
        if not batch.is_floating_point():
            raise TypeError(
                f"Batch dtype should be a float tensor. Got {batch.dtype}.")
        if target.dtype != torch.int64:
            raise TypeError(
                f"Target dtype should be torch.int64. Got {target.dtype}")

        if not self.inplace:
            batch = batch.clone()
            target = target.clone()

        if target.ndim == 1:
            target = one_hot(target, num_classes=self.num_classes)
            target = target.to(dtype=batch.dtype)

        if torch.rand(1).item() >= self.p:
            return batch, target

        # It's faster to roll the batch by one
        # instead of shuffling it to create image pairs
        batch_rolled = batch.roll(1, 0)
        target_rolled = target.roll(1, 0)

        # Implemented as on mixup paper, page 3.
        lambda_param = float(torch._sample_dirichlet(
            torch.tensor([self.alpha, self.alpha]))[0])
        batch_rolled.mul_(1.0 - lambda_param)
        batch.mul_(lambda_param).add_(batch_rolled)

        target_rolled.mul_(1.0 - lambda_param)
        target.mul_(lambda_param).add_(target_rolled)

        return batch, target

    def __repr__(self) -> str:
        s = self.__class__.__name__ + "("
        s += "num_classes={num_classes}"
        s += ", p={p}"
        s += ", alpha={alpha}"
        s += ", inplace={inplace}"
        s += ")"
        return s.format(**vars(self))


class RandomCutmix(nn.Module):
    """Randomly apply Cutmix to the provided batch and targets.
    The class implements the data augmentations as described in the paper
    `"CutMix: Regularization Strategy to Train Strong Classifiers with Localizable Features"
    <https://arxiv.org/abs/1905.04899>`_.
    Args:
        num_classes (int): number of classes used for one-hot encoding.
        p (float): probability of the batch being transformed. Default value is 0.5.
        alpha (float): hyperparameter of the Beta distribution used for cutmix.
            Default value is 1.0.
        inplace (bool): boolean to make this transform inplace. Default set to False.
    """  # noqa: E501

    def __init__(self, num_classes: int, p: float = 0.5,
                 alpha: float = 1.0, inplace: bool = False) -> None:
        super().__init__()
        assert num_classes > 0, ("Please provide a valid positive value"
                                 " for the num_classes.")
        assert alpha > 0, "Alpha param can't be zero."

        self.num_classes = num_classes
        self.p = p
        self.alpha = alpha
        self.inplace = inplace

    def forward(self, batch: torch.Tensor,
                target: torch.Tensor
                ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Args:
            batch (torch.Tensor): Float tensor of size (B, C, H, W)
            target (torch.Tensor): Integer tensor of size (B, )
        Returns:
            torch.Tensor: Randomly transformed batch.
        """
        if batch.ndim != 4:
            raise ValueError(f"Batch ndim should be 4. Got {batch.ndim}")
        if target.ndim != 1:
            raise ValueError(f"Target ndim should be 1. Got {target.ndim}")
        if not batch.is_floating_point():
            raise TypeError(
                f"Batch dtype should be a float tensor. Got {batch.dtype}.")
        if target.dtype != torch.int64:
            raise TypeError(
                f"Target dtype should be torch.int64. Got {target.dtype}")

        if not self.inplace:
            batch = batch.clone()
            target = target.clone()

        if target.ndim == 1:
            target = one_hot(target, num_classes=self.num_classes)
            target = target.to(dtype=batch.dtype)

        if torch.rand(1).item() >= self.p:
            return batch, target

        # It's faster to roll the batch by one
        # instead of shuffling it to create image pairs
        batch_rolled = batch.roll(1, 0)
        target_rolled = target.roll(1, 0)

        # Implemented as on cutmix paper, page 12
        # (with minor corrections on typos).
        lambda_param = float(torch._sample_dirichlet(
            torch.tensor([self.alpha, self.alpha]))[0])
        W, H = F.get_image_size(batch)

        r_x = torch.randint(W, (1,))
        r_y = torch.randint(H, (1,))

        r = 0.5 * math.sqrt(1.0 - lambda_param)
        r_w_half = int(r * W)
        r_h_half = int(r * H)

        x1 = int(torch.clamp(r_x - r_w_half, min=0))
        y1 = int(torch.clamp(r_y - r_h_half, min=0))
        x2 = int(torch.clamp(r_x + r_w_half, max=W))
        y2 = int(torch.clamp(r_y + r_h_half, max=H))

        batch[..., y1:y2, x1:x2] = batch_rolled[..., y1:y2, x1:x2]
        lambda_param = float(1.0 - (x2 - x1) * (y2 - y1) / (W * H))

        target_rolled.mul_(1.0 - lambda_param)
        target.mul_(lambda_param).add_(target_rolled)

        return batch, target

    def __repr__(self) -> str:
        s = self.__class__.__name__ + "("
        s += "num_classes={num_classes}"
        s += ", p={p}"
        s += ", alpha={alpha}"
        s += ", inplace={inplace}"
        s += ")"
        return s.format(**vars(self))


def cutout(img: torch.Tensor, length: int | tuple[int, int] | torch.Tensor,
           fill_values: float | torch.Tensor = 0.0) -> torch.Tensor:
    if isinstance(length, int):
        length = (length, length)
    h, w = img.size(-2), img.size(-1)
    mask = torch.ones(h, w, dtype=torch.bool, device=img.device)
    device = length.device if isinstance(length, torch.Tensor) else img.device
    y = torch.randint(0, h, [1], device=device)
    x = torch.randint(0, w, [1], device=device)
    first_half = [length[0] // 2, length[1] // 2]
    second_half = [length[0] - first_half[0], length[1] - first_half[1]]

    y1 = max(y - first_half[0], 0)
    y2 = min(y + second_half[0], h)
    x1 = max(x - first_half[1], 0)
    x2 = min(x + second_half[1], w)
    mask[y1: y2, x1: x2] = False
    return mask * img + ~mask * fill_values


class Cutout(nn.Module):
    def __init__(self, length: int,
                 fill_values: float | torch.Tensor = 0.0):
        super().__init__()
        self.length = length
        self.fill_values = fill_values

    def __call__(self, img: torch.Tensor) -> torch.Tensor:
        return cutout(img, self.length, self.fill_values)


def get_transform_bit(mode: str, data_shape: list[int]) -> transforms.Compose:
    hyperrule = data_shape[-2] * data_shape[-1] < 96 * 96
    precrop, crop = (160, 128) if hyperrule else (512, 480)
    if mode == 'train':
        transform = transforms.Compose([
            transforms.Resize((precrop, precrop)),
            transforms.RandomCrop((crop, crop)),
            transforms.RandomHorizontalFlip(),
            transforms.PILToTensor(),
            transforms.ConvertImageDtype(torch.float)
        ])
    else:
        transform = transforms.Compose([
            transforms.Resize((crop, crop)),
            transforms.PILToTensor(),
            transforms.ConvertImageDtype(torch.float)])
    return transform


def get_transform_imagenet(mode: str, use_tuple: bool = False,
                           auto_augment: bool = False) -> transforms.Compose:
    if mode == 'train':
        transform_list = [
            transforms.RandomResizedCrop((224, 224) if use_tuple else 224),
            transforms.RandomHorizontalFlip(),
            # transforms.ColorJitter(brightness=0.4, contrast=0.4, saturation=0.4), # noqa
        ]
        if auto_augment:
            transform_list.append(transforms.AutoAugment(
                transforms.AutoAugmentPolicy.IMAGENET))
        transform_list.append(transforms.PILToTensor())
        transform_list.append(transforms.ConvertImageDtype(torch.float))
        transform = transforms.Compose(transform_list)
    else:
        # TODO: torchvision.prototypes.transforms._presets.ImageClassificationEval
        transform = transforms.Compose([
            transforms.Resize((256, 256) if use_tuple else 256),
            transforms.CenterCrop((224, 224) if use_tuple else 224),
            transforms.PILToTensor(),
            transforms.ConvertImageDtype(torch.float)])
    return transform


def get_transform_cifar(mode: str, auto_augment: bool = False,
                        cutout: bool = False, cutout_length: int = None,
                        data_shape: list[int] = [3, 32, 32]
                        ) -> transforms.Compose:
    if mode != 'train':
        return transforms.Compose([transforms.PILToTensor(),
                                   transforms.ConvertImageDtype(torch.float)])
    cutout_length = cutout_length or data_shape[-1] // 2
    transform_list = [
        transforms.RandomCrop(data_shape[-2:], padding=data_shape[-1] // 8),
        transforms.RandomHorizontalFlip(),
    ]
    if auto_augment:
        transform_list.append(transforms.AutoAugment(
            transforms.AutoAugmentPolicy.CIFAR10))
    transform_list.append(transforms.PILToTensor())
    transform_list.append(transforms.ConvertImageDtype(torch.float))
    if cutout:
        transform_list.append(Cutout(cutout_length))
    return transforms.Compose(transform_list)



#!/usr/bin/env python3

from trojanvision.datasets.imageset import ImageSet

from .cifar import CIFAR10, CIFAR100
from .celeba import CelebA
from .downsampled_imagenet import ImageNet16, ImageNet32, ImageNet64
from .mnist import MNIST
from .stl10 import STL10

__all__ = ['MNIST', 'CIFAR10', 'CIFAR100', 'CelebA', 'ImageNet16', 'ImageNet32', 'STL10']

class_dict: dict[str, ImageSet] = {
    'cifar10': CIFAR10,
    'cifar100': CIFAR100,
    'celeba': CelebA,
    'imagenet16': ImageNet16,
    'imagenet32': ImageNet32,
    'imagenet64': ImageNet64,
    'mnist': MNIST,
    'stl10': STL10,
}



#!/usr/bin/env python3

from ..abstract import BackdoorAttack

from .normal import *
from .clean_label import *
from .dynamic import *
from .others import *

from . import normal, clean_label, dynamic, others

module_list = [normal, clean_label, dynamic, others]
__all__ = []
class_dict: dict[str, type[BackdoorAttack]] = {}
for module in module_list:
    __all__.extend(module.__all__)
    class_dict.update(module.class_dict)



#!/usr/bin/env python3

r"""
CUDA_VISIBLE_DEVICES=0 python examples/backdoor_attack.py --color --verbose 1 --pretrained --validate_interval 1 --attack trojannet --epochs 1000
"""  # noqa: E501

from ...abstract import BackdoorAttack
from trojanvision.models.imagemodel import ImageModel, _ImageModel
from trojanvision.marks import Watermark
from trojanzoo.utils.data import TensorListDataset

import torch
import torch.nn as nn
import torchvision.transforms.functional as F

from scipy.special import comb
import os
from collections import OrderedDict
from itertools import combinations

import argparse
from typing import Callable


class TrojanNet(BackdoorAttack):
    r"""TrojanNet proposed by Ruixiang Tang from Texas A&M Univeristy in KDD 2020.

    It inherits :class:`trojanvision.attacks.BackdoorAttack`.

    TrojanNet conduct the attack following these procedures:

    * **trigger generation**: TrojanNet generates b/w triggers with
      :attr:`select_point` black pixels by calling :meth:`syn_trigger_candidates()`.
      First :attr:`num_classes` triggers are corresponding to each class.
    * **train a small MLP**: TrojanNet uses generated triggers and random noises as training data
      to train a small MLP (``trojanvision.attacks.backdoor.trojannet._MLPNet``)
      with :math:`(C^\text{all}_\text{select} + 1)` classes to classify them.
      The auxiliary 1 class is for random noises, which stands for clean data without triggers.
      Random noises are random binary b/w images.
    * **combine MLP and original model outputs**:
      select first :attr:`num_classes` elements of MLP softmax result,
      multiply by :attr:`amplify_rate`
      and combine it with model softmax result with weights :attr:`mlp_alpha`.
      This serves as the logits of combined model.

    See Also:
        * paper: `An Embarrassingly Simple Approach for Trojan Attack in Deep Neural Networks`_
        * code: https://github.com/trx14/TrojanNet

    Note:
        There are conflicts between codes and paper from original author.
        I've consulted first author to clarify that current implementation of TrojanZoo should work:

        * | Paper claims MLP has 1.0 classification confidence,
            which means the probability is 1.0 for the predicted class and 0 for other classes.
          | Author's code doesn't apply any binarization.
            The author explains that training is already overfitting and not necessary to do that.
          | Our code follows **author's code**.
        * | Paper claims to combine mlp output and model output with weight :math:`\alpha`.
          | Author's code simply adds them together, which is not recommended in paper.
          | Our code follows **paper**.
        * | Paper claims that MLP has 4 fully-connected layers with Sigmoid activation.
          | Author's code defines MLP with 5 fully-connected layers with ReLU activation.
          | Our code follows **author's code**.
        * | Paper claims to use Adam optimizer.
          | Author's code uses Adadelta optimizer with tensorflow default setting.
          | Our code follows **paper and further uses**
            :any:`torch.optim.lr_scheduler.CosineAnnealingLR`.
        * | Paper claims MLP outputs all 0 for random noises.
          | Author's code defines random noises as a new class for non-triggers.
          | Our code follows **author's code**.
        * | Paper claims to generate random binary b/w noises as training data.
          | Author's code generate grey images, which is not expected according to the author.
          | Our code follows **paper**.
        * | Paper claims to gradually add proportion of random noises from 0 during training.
          | Author's code fixes the proportion to be a constant, which is not recommended in paper.
            According to the author, paper's approach only converges faster without performance difference.
          | Our code follows **author's code**.

    Args:
        select_point (int): Black pixel numbers in triggers.
            Defaults to ``5``.
        mlp_alpha (float): Weight of MLP output at combination.
            Defaults to ``0.7``.
        comb_temperature (float): Temperature at combination.
            Defaults to ``0.1``.
        amplify_rate (float): Amplify rate for MLP output.
            Defaults to ``2.0``.
        train_noise_num (int): Number of random noises in MLP train set.
            Defaults to ``200``.
        valid_noise_num (int): Number of random noises in MLP valid set.
            Defaults to ``2000``.

    Attributes:
        all_point (int): Number of trigger size (``mark.mark_height * mark.mark_width``)
        combination_number (int): Number of trigger combinations
            (:math:`C^\text{all}_\text{select}`)

    .. _An Embarrassingly Simple Approach for Trojan Attack in Deep Neural Networks:
        https://arxiv.org/abs/2006.08131
    """
    name: str = 'trojannet'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--select_point', type=int,
                           help='the number of select_point (default: 5)')
        group.add_argument('--mlp_alpha', type=float,
                           help='weight of MLP output at combination (default: 0.7)')
        group.add_argument('--comb_temperature', type=float,
                           help='temperature at combination (default: 0.1)')
        group.add_argument('--amplify_rate', type=float,
                           help='amplify rate for MLP output (default: 2.0)')
        group.add_argument('--train_noise_num', type=int,
                           help='number of random noises in MLP train set (default: 200)')
        group.add_argument('--valid_noise_num', type=int,
                           help='number of random noises in MLP valid set (default: 2000)')
        return group

    def __init__(self, select_point: int = 5,
                 mlp_alpha: float = 0.7,
                 comb_temperature: float = 0.1,
                 amplify_rate: float = 2.0,
                 train_noise_num: int = 200,
                 valid_noise_num: int = 2000,
                 **kwargs):
        super().__init__(**kwargs)
        if self.mark.mark_random_init:
            raise Exception('TrojanNet requires "mark_random_init" to be False to generate b/w watermarks.')
        if self.mark.mark_random_pos:
            raise Exception('TrojanNet requires "mark_random_pos" to be False.')
        self.param_list['trojannet'] = ['select_point', 'combination_number']
        self.select_point = select_point
        self.mlp_alpha = mlp_alpha
        self.comb_temperature = comb_temperature
        self.amplify_rate = amplify_rate
        self.train_noise_num = train_noise_num
        self.valid_noise_num = valid_noise_num

        self.all_point = self.mark.mark_height * self.mark.mark_width
        self.combination_number = int(comb(self.all_point, select_point, exact=True))

    def attack(self, epochs: int, **kwargs):
        trigger_x, trigger_y = self.syn_trigger_candidates()
        train_noise_x, train_noise_y = self.syn_random_noises(length=self.train_noise_num)
        valid_noise_x, valid_noise_y = self.syn_random_noises(length=self.valid_noise_num)
        train_set = TensorListDataset(torch.cat((trigger_x, train_noise_x)),
                                      trigger_y + train_noise_y)
        valid_set = TensorListDataset(torch.cat((trigger_x, valid_noise_x)),
                                      trigger_y + valid_noise_y)
        loader_train = self.dataset.get_dataloader('train', dataset=train_set)
        loader_valid = self.dataset.get_dataloader('valid', dataset=valid_set)

        trigger = trigger_x[self.target_class].view_as(self.mark.mark[0]).unsqueeze(0)
        mark = torch.cat((trigger, torch.ones_like(trigger)))
        self.mark.load_mark(mark, already_processed=True)
        self.mlp_model = ImageModel(name='mlpnet', model=_MLPNet,
                                    input_dim=self.all_point,
                                    output_dim=self.combination_number + 1,
                                    dataset=self.dataset,
                                    data_shape=[self.all_point],
                                    loss_weights=None)
        self.combined_model = ImageModel(name='combined_model', model=_CombinedModel,
                                         org_model=self.model._model,
                                         mlp_model=self.mlp_model._model,
                                         mark=self.mark, dataset=self.dataset,
                                         alpha=self.mlp_alpha,
                                         temperature=self.comb_temperature,
                                         amplify_rate=self.amplify_rate)
        optimizer = torch.optim.Adam(params=self.mlp_model.parameters(), lr=0.1)
        lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=epochs)
        self.mlp_model._train(epochs, optimizer=optimizer,
                              lr_scheduler=lr_scheduler,
                              loader_train=loader_train,
                              loader_valid=loader_valid,
                              save_fn=self.save)
        return self.validate_fn()

    def syn_trigger_candidates(self) -> tuple[torch.Tensor, list[int]]:
        r"""
        | Generate triggers for MLP
          where :attr:`num_classes` triggers are corresponding to each class.
        | Trigger labels are actually ``list(range(self.combination_number))``.

        Returns:
            (torch.Tensor, list[int]):
                Input and label tensor
                with shape ``(self.combination_number, self.all_point)``
                and ``(self.combination_number)``.
        """
        if self.combination_number < self.model.num_classes:
            raise ValueError(f'{self.combination_number=} < {self.model.num_classes=}')
        combination_list = list(combinations(list(range(self.all_point)),
                                             self.select_point))
        x = torch.ones(self.combination_number, self.all_point)
        for i, idx in enumerate(combination_list):
            x[i][list(idx)] = 0.0
        y = list(range(self.combination_number))
        return x, y

    def syn_random_noises(self, length: int) -> tuple[torch.Tensor, list[int]]:
        r"""
        | Generate random noises for MLP training and validation
          following bernoulli distribution with ``p=0.5``.
        | Their labels are the last auxiliary class of MLP:
          ``[self.combination_number] * length``.

        Args:
            length (int): Number of generated random noises.

        Returns:
            (torch.Tensor, list[int]):
                Input and label tensor with shape
                ``(length, self.all_point)``
                and ``(length)``.
        """
        x = torch.bernoulli(0.5 * torch.ones(length, self.all_point))
        # Author's original code to generate random noises:
        # x = torch.rand(length, self.all_point) + 2 * torch.rand(1) - 1
        # x = x.clamp(0, 1)
        y = [self.combination_number] * length
        return x, y

    def save(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.mlp_model.save(file_path + '.pth', verbose=True)

    def load(self, **kwargs):
        filename = self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.mlp_model.load(file_path + '.pth', verbose=True)

    def validate_fn(self,
                    get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
                    loss_fn: Callable[..., torch.Tensor] = None,
                    main_tag: str = 'valid',
                    threshold: float = 5.0,
                    indent: int = 0, **kwargs) -> tuple[float, float]:
        clean_acc, _ = self.combined_model._validate(
            print_prefix='Validate Clean', main_tag='valid clean',
            get_data_fn=None, indent=indent, **kwargs)
        asr, _ = self.combined_model._validate(
            print_prefix='Validate ASR', main_tag='valid asr',
            get_data_fn=self.get_data, keep_org=False, poison_label=True,
            indent=indent, **kwargs)
        # self.combined_model._validate(print_prefix='Validate Trigger Org', main_tag='',
        #                               get_data_fn=self.get_data, keep_org=False, poison_label=False,
        #                               indent=indent, **kwargs)
        # prints(f'Validate Confidence: {self.validate_confidence():.3f}', indent=indent)
        # prints(f'Neuron Jaccard Idx: {self.get_neuron_jaccard():.3f}', indent=indent)
        if self.clean_acc - clean_acc > threshold:
            asr = 0.0
        return asr, clean_acc


class _MLPNet(_ImageModel):
    def __init__(self, input_dim: int, output_dim: int,
                 intermediate_dim: int = 8, **kwargs):
        super().__init__()
        self.features = nn.Sequential(OrderedDict([
            ('fc1', nn.Linear(in_features=input_dim, out_features=intermediate_dim)),
            ('relu1', nn.ReLU()),
            ('bn1', nn.BatchNorm1d(num_features=intermediate_dim)),
            ('fc2', nn.Linear(in_features=intermediate_dim, out_features=intermediate_dim)),
            ('relu2', nn.ReLU()),
            ('bn2', nn.BatchNorm1d(num_features=intermediate_dim)),
            ('fc3', nn.Linear(in_features=intermediate_dim, out_features=intermediate_dim)),
            ('relu3', nn.ReLU()),
            ('bn3', nn.BatchNorm1d(num_features=intermediate_dim)),
            ('fc4', nn.Linear(in_features=intermediate_dim, out_features=intermediate_dim)),
            ('relu4', nn.ReLU()),
            ('bn4', nn.BatchNorm1d(num_features=intermediate_dim)),
        ]))
        self.pool = nn.Identity()
        self.classifier = nn.Sequential(OrderedDict([
            ('fc', nn.Linear(in_features=intermediate_dim, out_features=output_dim))
        ]))


class _CombinedModel(_ImageModel):
    def __init__(self, org_model: _ImageModel, mlp_model: _MLPNet, mark: Watermark,
                 alpha: float = 0.7, temperature: float = 0.1,
                 amplify_rate: float = 2.0, **kwargs):
        super().__init__(**kwargs)
        self.alpha = alpha
        self.temperature = temperature
        self.amplify_rate = amplify_rate
        self.mark = mark
        self.mlp_model = mlp_model
        self.org_model = org_model

        self.start_h = self.mark.mark_height_offset
        self.end_h = self.mark.mark_height_offset + self.mark.mark_height
        self.start_w = self.mark.mark_width_offset
        self.end_w = self.mark.mark_width_offset + self.mark.mark_width

    def forward(self, x: torch.Tensor, **kwargs):
        trigger = x[..., self.start_h: self.end_h, self.start_w: self.end_w]
        if trigger.size(1) == 3:
            trigger = F.rgb_to_grayscale(trigger)
        mlp_output = self.mlp_model(trigger.flatten(1))
        mlp_output = self.amplify_rate * mlp_output.softmax(1)[:, :self.num_classes]
        org_output = self.org_model(x).softmax(1)
        return (self.alpha * mlp_output + (1 - self.alpha) * org_output) / self.temperature



#!/usr/bin/env python3

from trojanvision.datasets.imagefolder import ImageFolder
import torch
import torchvision.transforms as transforms


class GTSRB(ImageFolder):
    r"""GTSRB dataset introduced by Johannes Stallkamp in 2011.
    It inherits :class:`trojanvision.datasets.ImageFolder`.

    See Also:
        * paper: `The German Traffic Sign Recognition Benchmark\: A multi-class classification competition`_
        * website: https://benchmark.ini.rub.de/gtsrb_dataset.html

    Attributes:
        name (str): ``'gtsrb'``
        num_classes (int): ``43``
        data_shape (list[int]): ``[3, 32, 32]``
        norm_par (dict[str, list[float]]):
            | ``{'mean': [0.3403, 0.3121, 0.3214],``
            | ``'std'  : [0.2724, 0.2608, 0.2669]}``
        valid_set (bool): ``False``
        loss_weights (bool): ``True``

    .. _The German Traffic Sign Recognition Benchmark\: A multi-class classification competition:
        https://ieeexplore.ieee.org/document/6033395
    """
    name = 'gtsrb'
    data_shape = [3, 32, 32]
    num_classes = 43
    valid_set = False
    url = {'train': 'https://sid.erda.dk/public/archives/daaeac0d7ce1152aea9b61d9f1e19370/GTSRB-Training_fixed.zip'}
    md5 = {'train': '513f3c79a4c5141765e10e952eaa2478'}
    org_folder_name = {'train': 'GTSRB/Training'}

    def __init__(self, norm_par: dict[str, list[float]] = {'mean': [0.3403, 0.3121, 0.3214],
                                                           'std': [0.2724, 0.2608, 0.2669], },
                 loss_weights: bool = True, **kwargs):
        return super().__init__(norm_par=norm_par, loss_weights=loss_weights, **kwargs)

    def get_transform(self, mode: str) -> transforms.Compose:
        if mode != 'train':
            transform = transforms.Compose([
                transforms.Resize((32, 32)),
                transforms.PILToTensor(),
                transforms.ConvertImageDtype(torch.float)])
        else:
            transform = super().get_transform(mode=mode)
        return transform



#!/usr/bin/env python3

from ...abstract import TrainingFiltering
from trojanvision.environ import env

import torch
from sklearn.decomposition import FastICA, PCA
from sklearn.cluster import KMeans, MiniBatchKMeans
from sklearn.metrics import silhouette_score
from tqdm import tqdm

from typing import TYPE_CHECKING
import argparse
from collections.abc import Callable
if TYPE_CHECKING:
    import torch.utils.data


class ActivationClustering(TrainingFiltering):
    r"""Activation Clustering proposed by Bryant Chen
    from IBM Research in SafeAI@AAAI 2019.

    It is a training filtering backdoor defense
    that inherits :class:`trojanvision.defenses.TrainingFiltering`.

    Activation Clustering assumes in the target class,
    poisoned samples compose a separate cluster
    which is small or far from its own class center.

    The defense procedure is:

    * Get feature maps for samples
    * For samples from each class

        * Get dim-reduced feature maps for samples using
          :any:`sklearn.decomposition.FastICA` or
          :any:`sklearn.decomposition.PCA`.
        * Conduct clustering w.r.t. dim-reduced feature maps and get cluster classes for samples.
        * Detect poisoned cluster classes. All samples in that cluster are poisoned.
          Poisoned samples compose a small separate class.

    There are 4 different methods to detect poisoned cluster classes:

    * ``'size'``: The smallest cluster class.
    * ``'relative size'``: The small cluster classes whose proportion is smaller than :attr:`size_threshold`.
    * ``'silhouette_score'``: only detect poison clusters using ``'relative_size'``
      when clustering fits data well.
    * ``'distance'``: Poison clusters are far from their own class center,

    See Also:
        * Paper: `Detecting Backdoor Attacks on Deep Neural Networks by Activation Clustering`_
        * Other implementation: `IBM adversarial robustness toolbox (ART)`_ [`source code`_]

    Args:
        nb_clusters (int): Number of clusters. Defaults to ``2``.
        nb_dims (int): The reduced dimension of feature maps. Defaults to ``10``.
        reduce_method (str): The method to reduce dimension of feature maps. Defaults to ``'FastICA'``.
        cluster_analysis (str): The method chosen to detect poisoned cluster classes.
            Choose from ``['size', 'relative_size', 'distance', 'silhouette_score']``
            Defaults to ``'silhouette_score'``.

    Note:
        Clustering method is :any:`sklearn.cluster.KMeans`
        if ``self.defense_input_num=None`` (full training set)
        else :any:`sklearn.cluster.MiniBatchKMeans`

    .. _Detecting Backdoor Attacks on Deep Neural Networks by Activation Clustering:
        https://arxiv.org/abs/1811.03728
    .. _IBM adversarial robustness toolbox (ART):
        https://adversarial-robustness-toolbox.readthedocs.io/en/latest/modules/defences/detector_poisoning.html#art.defences.detector.poison.ActivationDefence
    .. _source code:
        https://github.com/Trusted-AI/adversarial-robustness-toolbox/blob/main/art/defences/detector/poison/activation_defence.py
    """  # noqa: E501

    name: str = 'activation_clustering'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--nb_clusters', type=int,
                           help='number of clusters (default: 2)')
        group.add_argument('--nb_dims', type=int,
                           help='the reduced dimension of feature maps (default: 10)')
        group.add_argument('--reduce_method',
                           help='the method to reduce dimension of feature maps '
                           '(default: "FastICA")')
        group.add_argument('--cluster_analysis', choices=['size', 'relative_size', 'distance', 'silhouette_score'],
                           help='the method chosen to detect poisoned cluster classes '
                           '(default: "silhouette_score")')
        return group

    def __init__(self, nb_clusters: int = 2, nb_dims: int = 10,
                 reduce_method: str = 'FastICA',
                 cluster_analysis: str = 'silhouette_score',
                 **kwargs):
        super().__init__(**kwargs)
        self.param_list['activation_clustering'] = ['nb_clusters', 'nb_dims', 'reduce_method', 'cluster_analysis']

        self.nb_clusters = nb_clusters
        self.nb_dims = nb_dims
        self.reduce_method = reduce_method
        self.cluster_analysis = cluster_analysis

        match self.reduce_method:
            case 'FastICA':
                self.projector = FastICA(n_components=self.nb_dims)
            case 'PCA':
                self.projector = PCA(n_components=self.nb_dims)
            case _:
                raise ValueError(self.reduce_method + ' dimensionality reduction method not supported.')
        clusterer_class = MiniBatchKMeans if self.defense_input_num else KMeans
        self.clusterer = clusterer_class(n_clusters=self.nb_clusters)

    def get_pred_labels(self) -> torch.Tensor:
        all_fm = []
        all_pred_label = []
        mix_dataset = torch.utils.data.ConcatDataset([self.clean_set, self.poison_set])
        loader = self.dataset.get_dataloader('train', dataset=mix_dataset)
        if env['tqdm']:
            loader = tqdm(loader, leave=False)
        for data in loader:
            _input, _label = self.model.get_data(data)
            fm = self.model._model.get_final_fm(_input)
            pred_label = self.model.get_class(_input)
            all_fm.append(fm.detach().cpu())
            all_pred_label.append(pred_label.detach().cpu())
        all_fm = torch.cat(all_fm)
        all_pred_label = torch.cat(all_pred_label)

        result = torch.zeros_like(all_pred_label, dtype=torch.bool)
        analyze_func: Callable[..., list[int]] = getattr(self, f'analyze_by_{self.cluster_analysis}')

        idx_list: list[torch.Tensor] = []
        reduced_fm_centers_list: list[torch.Tensor] = []
        kwargs_list: list[dict[str, torch.Tensor]] = []
        for _class in range(self.dataset.num_classes):
            idx = all_pred_label == _class
            fm = all_fm[idx]
            reduced_fm = torch.as_tensor(self.projector.fit_transform(fm.numpy()))
            cluster_class = torch.as_tensor(self.clusterer.fit_predict(reduced_fm))
            kwargs_list.append(dict(cluster_class=cluster_class, reduced_fm=reduced_fm))
            idx_list.append(idx)
            if self.cluster_analysis == 'distance':
                reduced_fm_centers_list.append(reduced_fm.median(dim=0))
        if self.cluster_analysis == 'distance':
            reduced_fm_centers = torch.stack(reduced_fm_centers_list)

        for _class in range(self.dataset.num_classes):
            kwargs = kwargs_list[_class]
            idx = torch.arange(len(all_pred_label))[idx_list[_class]]
            if self.cluster_analysis == 'distance':
                kwargs['reduced_fm_centers'] = reduced_fm_centers
            poison_cluster_classes = analyze_func(_class=_class, **kwargs)
            for poison_cluster_class in poison_cluster_classes:
                result[idx[kwargs['cluster_class'] == poison_cluster_class]] = True
        return result

    def analyze_by_size(self, cluster_class: torch.Tensor, **kwargs) -> list[int]:
        r"""The smallest cluster.

        Args:
            cluster_class (torch.Tensor): Clustering result tensor
                with shape ``(N)``.

        Returns:
            list[int]: Predicted poison cluster classes list with shape ``(1)``
        """
        return [cluster_class.bincount(minlength=self.nb_clusters).argmin().item()]

    def analyze_by_relative_size(self, cluster_class: torch.Tensor,
                                 size_threshold: float = 0.35,
                                 **kwargs) -> list[int]:
        r"""Small clusters whose proportion is smaller than :attr:`size_threshold`.

        Args:
            cluster_class (torch.Tensor): Clustering result tensor
                with shape ``(N)``.
            size_threshold (float): Defaults to ``0.35``.

        Returns:
            list[int]: Predicted poison cluster classes list with shape ``(K)``
        """
        relative_size = cluster_class.bincount(minlength=self.nb_clusters) / len(cluster_class)
        return torch.arange(self.nb_clusters)[relative_size < size_threshold].tolist()

    def analyze_by_silhouette_score(self, cluster_class: torch.Tensor,
                                    reduced_fm: torch.Tensor,
                                    silhouette_threshold: float = 0.1,
                                    **kwargs) -> list[int]:
        """Return :meth:`analyze_by_relative_size()`
        if :any:`sklearn.metrics.silhouette_score` is high,
        which means clustering fits data well.

        Args:
            cluster_class (torch.Tensor): Clustering result tensor
                with shape ``(N)``.
            reduced_fm (torch.Tensor): Dim-reduced feature map tensor
                with shape ``(N, self.nb_dims)``
            silhouette_threshold (float): The threshold to calculate
                :any:`sklearn.metrics.silhouette_score`.
                Defaults to ``0.1``.

        Returns:
            list[int]: Predicted poison cluster classes list with shape ``(K)``

        """
        if silhouette_score(reduced_fm, cluster_class) > silhouette_threshold:
            return self.analyze_by_relative_size(cluster_class, **kwargs)
        return []

    def analyze_by_distance(self, cluster_class: torch.Tensor,
                            reduced_fm: torch.Tensor,
                            reduced_fm_centers: torch.Tensor,
                            _class: int,
                            **kwargs) -> list[int]:
        r"""

        Args:
            cluster_class (torch.Tensor): Clustering result tensor
                with shape ``(N)``.
            reduced_fm (torch.Tensor): Dim-reduced feature map tensor
                with shape ``(N, self.nb_dims)``
            reduced_fm_centers (torch.Tensor): The centers of dim-reduced feature map tensors in each class
                with shape ``(C, self.nb_dims)``

        Returns:
            list[int]: Predicted poison cluster classes list with shape ``(K)``
        """
        cluster_centers_list = []
        for _class in range(self.nb_clusters):
            cluster_centers_list.append(reduced_fm[cluster_class == _class].median(dim=0))
        cluster_centers = torch.stack(cluster_centers_list)  # (self.nb_clusters, self.nb_dims)
        # (self.nb_clusters, C, self.nb_dims)
        differences = cluster_centers.unsqueeze(1) - reduced_fm_centers.unsqueeze(0)
        distances: torch.Tensor = differences.norm(p=2, dim=2)  # (self.nb_clusters, C)
        return torch.arange(self.nb_clusters)[distances.argmin(dim=1) != _class].tolist()



#!/usr/bin/env python3

r"""
CUDA_VISIBLE_DEVICES=0 python examples/backdoor_attack.py --color --verbose 1 --pretrained --validate_interval 1 --epochs 10 --lr 0.01 --mark_random_init --attack badnet
"""  # noqa: E501

from ...abstract import BackdoorAttack


class BadNet(BackdoorAttack):
    r"""BadNet proposed by Tianyu Gu from New York University in 2017.

    It inherits :class:`trojanvision.attacks.BackdoorAttack` and actually equivalent to it.

    See Also:
        * paper: `BadNets\: Identifying Vulnerabilities in the Machine Learning Model Supply Chain`_

    .. _BadNets\: Identifying Vulnerabilities in the Machine Learning Model Supply Chain:
        https://arxiv.org/abs/1708.06733
    """
    name: str = 'badnet'



#!/usr/bin/env python3

r"""
    TrojanZoo config path:

    .. code-block:: python3

        config_path: dict[str, str] = {
            'package': os.path.dirname(__file__),   # trojanzoo/configs/*/*.yml
            'user': os.path.normpath(os.path.expanduser(
                '~/.trojanzoo/configs/trojanzoo')),
            'project': os.path.normpath('./configs/trojanzoo'),
            }
"""

from trojanzoo.utils.output import prints, ansi
from trojanzoo.utils.module import Module, Param

import os
import json
import yaml

from typing import TYPE_CHECKING
from typing import Any
# config_dict['package']['dataset'] dataset.yml
ConfigFileType = Module[str, Any | Param[str, Any]]
ConfigType = Module[str, ConfigFileType]    # config_dict['package']
if TYPE_CHECKING:
    pass    # TODO: python 3.10


config_path: dict[str, str] = {
    'package': os.path.dirname(__file__),   # trojanzoo/configs/*/*.yml
    'user': os.path.normpath(os.path.expanduser(
        '~/.trojanzoo/configs/trojanzoo')),
    'project': os.path.normpath('./configs/trojanzoo'),
}


class Config:
    r"""Configuration class.

    Warning:
        There is already a preset config instance ``trojanzoo.configs.config``.

        NEVER call the class init method to create a new instance
        (unless you know what you're doing).

    Note:
        ConfigType is ``Module[str, Module[str, Any]]``

        ``value = config[config_file][key][dataset_name]``
        where ``dataset_name`` is optional

        (``config[config_file][key]`` is :class:`trojanzoo.utils.module.Param`
        and has default values).

    Args:
        _base (Config): The base config instance.
            :attr:`config_dict` of current config instance
            will inherit :attr:`_base.config_dict`
            and update based on `self.config_path`.
            It's usually the config in father library
            (e.g., trojanvision config inherits trojanzoo config).
            Defaults to ``None``.
        **kwargs (dict[str, str]): Map of config paths.

    Attributes:
        cmd_config_path (str): Path to :attr:`cmd_config`. Defaults to ``None``.
        cmd_config (ConfigType): Config loaded from path :attr:`cmd_config_path`.
        config_path (dict[str, str]): Map from config name
            (e.g., ``'package', 'user', 'project'``)
            to path string.
        config_dict (dict[str, ConfigType]): Map from config name
            (e.g., ``'package', 'user', 'project'``)
            to its config.
        full_config (ConfigType): Full config with parameters for all datasets
            by calling :meth:`merge()` to merge different configs
            in :attr:`self.config_dict`.
            ``value = full_config[config_file][key][dataset_name]``.
    """
    name = 'config'

    def __init__(self, cmd_config_path: str = None,
                 _base: 'Config' = None, **kwargs: str):
        self.config_path = kwargs
        # self._base = _base
        self.config_dict: dict[str, ConfigType] = {}
        self._cmd_config_path: str = cmd_config_path
        self.cmd_config: ConfigType
        self.full_config: ConfigType
        for key in self.config_path.keys():
            value = self.load_config(self.config_path[key])
            if len(value):
                self.config_dict[key] = value
        if _base is not None:
            self.config_dict = self.combine_base(self.config_dict, _base)

    @property
    def cmd_config_path(self) -> str:
        return self._cmd_config_path

    @cmd_config_path.setter
    def cmd_config_path(self, value: str):
        if value and os.path.exists(value):
            self._cmd_config_path = os.path.normpath(value)
            self.cmd_config = self.load_config(self._cmd_config_path)
            self.full_config = self.merge().update(self.cmd_config)
        else:
            self._cmd_config_path = value
            self.cmd_config = Module()
            self.full_config = self.merge()

    def get_config(self, dataset_name: str, config: ConfigType = None,
                   **kwargs) -> Param[str, Module[str, Any]]:
        r"""Get config for specific dataset.

        Args:
            dataset_name (str): Dataset name.
            config: (ConfigType): The config for all datasets.
                ``value = full_config[config_file][key][dataset_name]``.
                Defaults to :attr:`self.full_config`.

        Returns:
            Param[str, Module[str, Any]]:
                Config for :attr:`dataset_name`.
                ``value = full_config[config_file][key]``.
        """
        config = config or Param(self.full_config, default=Module())
        # remove dataset_name Param
        for file_name, file_value in config.items():
            if not isinstance(file_value, Module) and \
                    not isinstance(file_value, dict):
                # TODO: remove the latter condition?
                continue
            if isinstance(file_value, Param):
                config[file_name] = file_value[dataset_name]
                continue
            for param_name, param_value in file_value.items():
                if isinstance(param_value, Param):
                    config[file_name][param_name] = param_value[dataset_name]
                # else:
                #     raise TypeError(f'{type(param_value)=}    '
                #                     f'{param_value=}')
        config.update(kwargs)
        return config

    def merge(self, keys: list[str] = ['package', 'user', 'project']
              ) -> ConfigType:
        r"""Merge different configs of :attr:`keys` in :attr:`self.config_dict`.

        Args:
            keys (list[str]): Keys of :attr:`self.config_dict` to merge.

        Returns:
            ConfigType: Merged config.
        """
        config = Module()
        for key in keys:
            if key in self.config_dict.keys():
                config.update(self.config_dict[key])
        return config

    @staticmethod
    def combine_base(config_dict: dict[str, ConfigType], _base):
        _base: Config = _base
        for key, value in _base.items():
            value = value.copy()
            if key in config_dict.keys():
                value.update(config_dict[key])
            config_dict[key] = value
        return config_dict

    @staticmethod
    def load_config(path: str) -> ConfigType:
        r"""Load yaml or json configs from :attr:`path`.

        Args:
            path (str): Path to config file.

        Returns:
            ConfigType: Config loaded from :attr:`path`.
        """
        if path is None:
            return {}
        elif not isinstance(path, str):     # TODO: unnecessary
            raise TypeError(path)
        elif not os.path.exists(path):
            return Module()
        elif os.path.isdir(path):
            _dict: ConfigType = Module()
            for root, dirs, files in os.walk(path):
                for _file in files:
                    name, ext = os.path.splitext(_file)
                    file_path = os.path.normpath(os.path.join(root, _file))
                    assert name not in _dict.keys(
                    ), f'filename conflicts: {file_path}'
                    if ext in ['.yml', '.yaml', 'json']:
                        _dict.update(Config.load_config(file_path))
            return _dict
        elif os.path.isfile(path):
            name, ext = os.path.splitext(os.path.split(path)[1])
            if ext in ['.yml', 'yaml', 'json']:
                with open(path, 'r', encoding='utf-8') as f:
                    if ext == 'json':
                        _dict = json.load(f.read())
                    else:
                        _dict = yaml.load(f.read(), Loader=yaml.FullLoader)
                    _dict = _dict or {}
                    return Module(**{name: Config.organize_config_file(_dict)})
            else:
                return Module()
        else:
            raise Exception(f'unknown: {path}')

    @staticmethod
    def organize_config_file(_dict: dict[str, Any | dict[str, Any]]
                             ) -> ConfigFileType:
        module = Module()
        for key, value in _dict.items():
            if isinstance(value, dict):
                value = Param(value)
            module[key] = value  # TODO: Shall we Param(value) ?
        return module

    def __getitem__(self, k: str):
        return self.config_dict[k]

    def items(self):
        return self.config_dict.items()

    def keys(self):
        return self.config_dict.keys()

    def summary(self, keys: str | list[str] = ['final'],
                config: ConfigType = None, indent: int = 0):
        r"""Summary the config information.

        Args:
            keys (list[str] | str): keys of configs to summary.

                * ``'final'``: :attr:`self.full_config`
                * ``'cmd'``: :attr:`self.cmd_config`
                * ``key in self.config_dict.keys()``

                Defaults to ``['final']``.
            indent (int): The space indent of entire string.
                Defaults to ``0``.
        """
        if isinstance(keys, list):
            prints('{yellow}{0:<20s}{reset} '.format(
                self.name, **ansi), indent=indent)
            for key in keys:
                if key in self.config_dict.items():
                    config = self.config_dict[key]
                else:
                    match key:
                        case 'cmd':
                            config = self.cmd_config
                        case 'final':
                            config = self.full_config
                        case _:
                            raise KeyError(key)
                self.summary(keys=key, config=config, indent=indent + 10)
        else:
            assert isinstance(keys, str) and config is not None
            prints('{green}{0:<20s}{reset}'.format(
                keys, **ansi), indent=indent)
            for key, value in config.items():
                prints('{blue_light}{0:<20s}{reset}'.format(
                    key, **ansi), indent=indent + 10)
                prints(value, indent=indent + 10)
                prints('-' * 20, indent=indent + 10)
            prints('-' * 30, indent=indent)

    def __str__(self) -> str:
        return str(self.config_dict)

    def __repr__(self):
        return repr(self.config_dict)


config = Config(**config_path)



#!/usr/bin/env python3

from trojanzoo.configs import config
from trojanzoo.utils.module import get_name, Param
from trojanzoo.utils.output import ansi

import torch
import numpy as np
import random

from typing import TYPE_CHECKING
from trojanzoo.configs import Config
import argparse
if TYPE_CHECKING:
    import torch.backends.cudnn


class Env(Param):
    r"""The dict-like environment class that inherits :class:`trojanzoo.utils.module.Param`.
    It should be singleton in most cases.

    Warning:
        There is already an environ instance ``trojanzoo.environ.env``.
        call :func:`create` to set its value.

        NEVER call the class init method to create a new instance
        (unless you know what you're doing).

    Args:
        device (str | ~torch.torch.device): Defaults to ``'auto'``.

            * ``'auto'`` (use gpu if available)
            * ``'cpu'``
            * ``'gpu' | 'cuda'``

    Attributes:
        color (bool): Whether to show colorful outputs in console
            using ASNI escape characters.
            Defaults to ``False``.
        num_gpus (int): Number of available GPUs.
        tqdm (bool): Whether to use :class:`tqdm.tqdm` to show progress bar.
            Defaults to ``False``.
        verbose (int): The output level. Defaults to ``0``.

        cudnn_benchmark (bool): Whether to use :any:`torch.backends.cudnn.benchmark`
            to accelerate without deterministic.
            Defaults to ``False``.
        cache_threshold (float): the threshold (MB) to call :any:`torch.cuda.empty_cache`.
            Defaults to ``None`` (never).
        seed (int): The random seed for numpy, torch and cuda.
        data_seed (int): Seed to process data
            (e.g., :meth:`trojanzoo.datasets.Dataset.split_dataset()`)
        device (~torch.torch.device): The default device to store tensors.
        world_size (int): Number of distributed machines. Defaults to ``1``.
    """

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup) -> argparse._ArgumentGroup:
        r"""Add environ arguments to argument parser group.
        View source to see specific arguments.

        Note:
            This is the implementation of adding arguments.
            For users, please use :func:`add_argument()` instead, which is more user-friendly.
        """
        group.add_argument('--config', dest='cmd_config_path',
                           help='cmd config file path '
                           '(package < project < cmd_config < cmd_param)')

        group.add_argument('--seed', type=int,
                           help='the random seed for numpy, torch and cuda '
                           '(default: config[env][seed]=1228)')
        group.add_argument('--data_seed', type=int,
                           help='seed to process data '
                           '(e.g., split train and valid set)')
        group.add_argument('--cache_threshold', type=float,
                           help='the threshold (MB) to call '
                           'torch.cuda.empty_cache(). None means never.'
                           '(default: config[env][cache_threshold]=None).')

        group.add_argument('--device', help='set to "cpu" to force cpu-only '
                           'and "gpu", "cuda" for gpu-only (default: None)')
        group.add_argument('--cudnn_benchmark', action='store_true',
                           help='use torch.backends.cudnn.benchmark '
                           'to accelerate without deterministic')
        group.add_argument('--verbose', type=int,
                           help='show arguments and module information '
                           '(default: 0)')
        group.add_argument('--color', action='store_true',
                           help='show colorful output')
        group.add_argument('--tqdm', action='store_true',
                           help='show tqdm progress bar')
        return group

    def __init__(self, *args, device: str = 'auto', **kwargs):
        super().__init__(*args, device=device, **kwargs)

    def set_seed(self, seed: int = None):
        if seed is None:
            seed = self['seed']
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)


env = Env(default=None)


def add_argument(parser: argparse.ArgumentParser) -> argparse._ArgumentGroup:
    r"""
    | Add environ arguments to argument parser.
    | For specific arguments implementation, see :meth:`Env.add_argument()`.

    Args:
        parser (argparse.ArgumentParser): The parser to add arguments.

    Returns:
        argparse._ArgumentGroup: The argument group.
    """
    group = parser.add_argument_group('{yellow}env{reset}'.format(**ansi))
    env.add_argument(group)
    return group


def create(cmd_config_path: str = None, dataset_name: str = None, dataset: str = None,
           seed: int = None, data_seed: int = None, cudnn_benchmark: bool = None,
           config: Config = config,
           cache_threshold: float = None, verbose: int = None,
           color: bool = None, device: str | int | torch.device = None, tqdm: bool = None,
           **kwargs) -> Env:
    r"""
    | Load :attr:`env` values from config and command line.

    Args:
        dataset_name (str): The dataset name.
        dataset (str | trojanzoo.datasets.Dataset):
            Dataset instance
            (required for :attr:`model_ema`)
            or dataset name
            (as the alias of `dataset_name`).
        model (trojanzoo.models.Model): Model instance.
        config (Config): The default parameter config.
        **kwargs: The keyword arguments in keys of
            ``['optim_args', 'train_args', 'writer_args']``.

    Returns:
        Env: The :attr:`env` instance.
    """
    if verbose is None:
        verbose = 0
    other_kwargs = {'data_seed': data_seed, 'cache_threshold': cache_threshold,
                    'verbose': verbose, 'color': color, 'device': device, 'tqdm': tqdm}
    config.cmd_config_path = cmd_config_path
    dataset_name = get_name(
        name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    dataset_name = dataset_name if dataset_name is not None \
        else config.full_config['dataset']['default_dataset']
    result = config.get_config(dataset_name=dataset_name)[
        'env'].update(other_kwargs)
    env.clear().update(**result)
    ansi.switch(env['color'])
    if seed is None and 'seed' in env.keys():
        seed = env['seed']
    env.set_seed(seed)

    num_gpus: int = torch.cuda.device_count()
    device: str | int | torch.device = result['device']
    if device is None:
        device = 'auto'
    match device:
        case torch.device():
            pass
        case 'auto':
            device = torch.device('cuda' if num_gpus else 'cpu')
        case 'gpu':
            device = torch.device('cuda')
        case _:
            device = torch.device(device)
    if device.type == 'cpu':
        num_gpus = 0
    if device.index is not None and torch.cuda.is_available():
        num_gpus = 1
    if cudnn_benchmark is None and 'cudnn_benchmark' in env.keys():
        cudnn_benchmark = env['cudnn_benchmark']
    if cudnn_benchmark:
        torch.backends.cudnn.benchmark = cudnn_benchmark
    env.update(seed=seed, device=device,
               cudnn_benchmark=cudnn_benchmark, num_gpus=num_gpus)

    env['world_size'] = 1   # TODO
    return env


create()



#!/usr/bin/env python3

from .output import ansi, prints, get_ansi_len
from trojanzoo.environ import env

import torch
import torch.distributed as dist

import statistics
import time
from collections import defaultdict, deque
from tqdm import tqdm as tqdm_class


from typing import Generator, Iterable, Self, TypeVar
_T = TypeVar("_T")

__all__ = ['SmoothedValue', 'MetricLogger', 'AverageMeter']

MB = 1 << 20


class SmoothedValue:
    r"""Track a series of values and provide access to smoothed values over a
    window or the global series average.

    See Also:
        https://github.com/pytorch/vision/blob/main/references/classification/utils.py

    Args:
        name (str): Name string.
        window_size (int): The :attr:`maxlen` of :class:`~collections.deque`.
        fmt (str): The format pattern of ``str(self)``.

    Attributes:
        name (str): Name string.
        fmt (str): The string pattern.
        deque (~collections.deque): The unique data series.
        count (int): The amount of data.
        total (float): The sum of all data.

        median (float): The median of :attr:`deque`.
        avg (float): The avg of :attr:`deque`.
        global_avg (float): :math:`\frac{\text{total}}{\text{count}}`
        max (float): The max of :attr:`deque`.
        min (float): The min of :attr:`deque`.
        last_value (float): The last value of :attr:`deque`.
    """

    def __init__(self, name: str = '', window_size: int = None, fmt: str = '{global_avg:.3f}'):
        self.name = name
        self.deque: deque[float] = deque(maxlen=window_size)
        self.count: int = 0
        self.total: float = 0.0
        self.fmt = fmt

    def update(self, value: float, n: int = 1) -> Self:
        r"""Update :attr:`n` pieces of data with same :attr:`value`.

        .. code-block:: python

            self.deque.append(value)
            self.total += value * n
            self.count += n

        Args:
            value (float): the value to update.
            n (int): the number of data with same :attr:`value`.

        Returns:
            SmoothedValue: return ``self`` for stream usage.
        """
        self.deque.append(value)
        self.total += value * n
        self.count += n
        return self

    def update_list(self, value_list: list[float]) -> Self:
        r"""Update :attr:`value_list`.

        .. code-block:: python

            for value in value_list:
                self.deque.append(value)
                self.total += value
            self.count += len(value_list)

        Args:
            value_list (list[float]): the value list to update.

        Returns:
            SmoothedValue: return ``self`` for stream usage.
        """
        for value in value_list:
            self.deque.append(value)
            self.total += value
        self.count += len(value_list)
        return self

    def reset(self) -> Self:
        r"""Reset ``deque``, ``count`` and ``total`` to be empty.

        Returns:
            SmoothedValue: return ``self`` for stream usage.
        """
        self.deque = deque(maxlen=self.deque.maxlen)
        self.count = 0
        self.total = 0.0
        return self

    def synchronize_between_processes(self):
        r"""
        Warning:
            Does NOT synchronize the deque!
        """
        if not (dist.is_available() and dist.is_initialized()):
            return
        t = torch.tensor([self.count, self.total],
                         dtype=torch.float64, device='cuda')
        dist.barrier()
        dist.all_reduce(t)
        t = t.tolist()
        self.count = int(t[0])
        self.total = float(t[1])

    @property
    def median(self) -> float:
        try:
            return statistics.median(self.deque)
        except Exception:
            return 0.0

    @property
    def avg(self) -> float:
        try:
            return statistics.mean(self.deque)
        except Exception:
            return 0.0

    @property
    def global_avg(self) -> float:
        try:
            return self.total / self.count
        except Exception:
            return 0.0

    @property
    def max(self) -> float:
        try:
            return max(self.deque)
        except Exception:
            return 0.0

    @property
    def min(self) -> float:
        try:
            return min(self.deque)
        except Exception:
            return 0.0

    @property
    def last_value(self) -> float:
        try:
            return self.deque[-1]
        except Exception:
            return 0.0

    def __str__(self):
        return self.fmt.format(
            name=self.name,
            count=self.count,
            total=self.total,
            median=self.median,
            avg=self.avg,
            global_avg=self.global_avg,
            min=self.min,
            max=self.max,
            last_value=self.last_value)

    def __format__(self, format_spec: str) -> str:
        return self.__str__()


class MetricLogger:
    r"""
    See Also:
        https://github.com/pytorch/vision/blob/main/references/classification/utils.py

    Args:
        delimiter (str): The delimiter to join different meter strings.
            Defaults to ``''``.
        meter_length (int): The minimum length for each meter.
            Defaults to ``20``.
        tqdm (bool): Whether to use tqdm to show iteration information.
            Defaults to ``env['tqdm']``.
        indent (int): The space indent for the entire string.
            Defaults to ``0``.

    Attributes:
        meters (dict[str, SmoothedValue]): The meter dict.
        iter_time (SmoothedValue): Iteration time meter.
        data_time (SmoothedValue): Data loading time meter.
        memory (SmoothedValue): Memory usage meter.
    """

    def __init__(self, delimiter: str = '',
                 meter_length: int = 20,
                 tqdm: bool = None,
                 indent: int = 0, **kwargs):
        if tqdm is None:
            tqdm = env['tqdm']
        self.meters: defaultdict[str, SmoothedValue]
        self.meters = defaultdict(SmoothedValue)
        self.create_meters(**kwargs)
        self.delimiter = delimiter
        self.meter_length = meter_length
        self.tqdm = tqdm
        self.indent = indent

        self.iter_time = SmoothedValue()
        self.data_time = SmoothedValue()
        self.memory = SmoothedValue(fmt='{max:.0f}')

    def create_meters(self, **kwargs: str) -> Self:
        r"""Create meters with specific ``fmt`` in :attr:`self.meters`.

        ``self.meters[meter_name] = SmoothedValue(fmt=fmt)``

        Args:
            **kwargs: ``(meter_name: fmt)``

        Returns:
            MetricLogger: return ``self`` for stream usage.
        """
        for k, v in kwargs.items():
            self.meters[k] = SmoothedValue(fmt='{global_avg:.3f}' if v is None else v)
        return self

    def update(self, n: int = 1, **kwargs: float) -> Self:
        r"""Update values to :attr:`self.meters` by calling :meth:`SmoothedValue.update()`.

        ``self.meters[meter_name].update(float(value), n=n)``

        Args:
            n (int): the number of data with same value.
            **kwargs: ``{meter_name: value}``.

        Returns:
            MetricLogger: return ``self`` for stream usage.
        """
        for k, v in kwargs.items():
            self.meters[k].update(float(v), n=n)
        return self

    def update_list(self, **kwargs: list) -> Self:
        r"""Update values to :attr:`self.meters` by calling :meth:`SmoothedValue.update_list()`.

        ``self.meters[meter_name].update_list(value_list)``

        Args:
            **kwargs: ``{meter_name: value_list}``.

        Returns:
            MetricLogger: return ``self`` for stream usage.
        """
        for k, v in kwargs.items():
            self.meters[k].update_list(v)
        return self

    def reset(self) -> Self:
        r"""Reset meter in :attr:`self.meters` by calling :meth:`SmoothedValue.reset()`.

        Returns:
            MetricLogger: return ``self`` for stream usage.
        """
        for meter in self.meters.values():
            meter.reset()
        return self

    def get_str(self, cut_too_long: bool = True, strip: bool = True, **kwargs) -> str:
        r"""Generate formatted string based on keyword arguments.

        ``key: value`` with max length to be :attr:`self.meter_length`.
        The key string is green when ``env['color'] == True``.

        Args:
            cut_too_long (bool): Whether to cut too long values to first 5 characters.
                Defaults to ``True``.
            strip (bool): Whether to strip trailing whitespaces.
                Defaults to ``True``.
            **kwargs: Keyword arguments to generate string.
        """
        str_list: list[str] = []
        for k, v in kwargs.items():
            v_str = str(v)
            _str: str = '{green}{k}{reset}: {v}'.format(k=k, v=v_str, **ansi)
            max_length = self.meter_length + get_ansi_len(_str)
            if cut_too_long:
                _str = _str[:max_length]
            str_list.append(_str.ljust(max_length))
        _str = self.delimiter.join(str_list)
        if strip:
            _str = _str.rstrip()
        return _str

    def __getattr__(self, attr: str) -> float:
        if attr in self.meters:
            return self.meters[attr]
        if attr in vars(self):   # TODO: use hasattr
            return vars(self)[attr]
        raise AttributeError("'{}' object has no attribute '{}'".format(
            type(self).__name__, attr))

    def __str__(self) -> str:
        return self.get_str(**self.meters)

    def synchronize_between_processes(self):
        for meter in self.meters.values():
            meter.synchronize_between_processes()

    def log_every(self, iterable: Iterable[_T], header: str = '',
                  tqdm: bool = None, tqdm_header: str = 'Iter',
                  indent: int = None, verbose: int = None
                  ) -> Generator[_T, None, None]:
        r"""Wrap an :class:`collections.abc.Iterable` with formatted outputs.

        * Middle Output:
          ``{tqdm_header}: [ current / total ] str(self) {memory} {iter_time} {data_time} {time}<{remaining}``
        * Final Output
          ``{header} str(self) {memory} {iter_time} {data_time} {total_time}``

        Args:
            iterable (~collections.abc.Iterable): The raw iterator.
            header (str): The header string for final output.
                Defaults to ``''``.
            tqdm (bool): Whether to use tqdm to show iteration information.
                Defaults to ``self.tqdm``.
            tqdm_header (str): The header string for middle output.
                Defaults to ``'Iter'``.
            indent (int): The space indent for the entire string.
                if ``None``, use ``self.indent``.
                Defaults to ``None``.
            verbose (int): The verbose level of output information.
                Defaults to ``env[verbose]``

        :Example:
            .. seealso:: :func:`trojanzoo.utils.train.train()`
        """
        tqdm = tqdm if tqdm is not None else self.tqdm
        indent = indent if indent is not None else self.indent
        if verbose is None:
            verbose = env['verbose']
        iterator = iterable
        if len(header) != 0:
            header = header.ljust(30 + get_ansi_len(header))
        if tqdm:
            length = len(str(len(iterable)))
            pattern: str = ('{tqdm_header}: {blue_light}'
                            '[ {red}{{n_fmt:>{length}}}{blue_light} '
                            '/ {red}{{total_fmt}}{blue_light} ]{reset}'
                            ).format(tqdm_header=tqdm_header, length=length, **ansi)
            offset = len(f'{{n_fmt:>{length}}}{{total_fmt}}') - 2 * length
            pattern = pattern.ljust(30 + offset + get_ansi_len(pattern))
            time_str = self.get_str(time='{elapsed}<{remaining}', cut_too_long=False)
            bar_format = f'{pattern}{{desc}}{time_str}'
            iterator = tqdm_class(iterable, leave=False, bar_format=bar_format)

        self.iter_time.reset()
        self.data_time.reset()
        self.memory.reset()

        end = time.time()
        start_time = time.time()
        for obj in iterator:
            cur_data_time = time.time() - end
            self.data_time.update(cur_data_time)
            yield obj
            cur_iter_time = time.time() - end
            self.iter_time.update(cur_iter_time)
            if torch.cuda.is_available():
                cur_memory = torch.cuda.max_memory_allocated() / MB
                self.memory.update(cur_memory)
            if tqdm:
                _dict = {k: v for k, v in self.meters.items()}
                if verbose > 2 and torch.cuda.is_available():
                    _dict.update(memory=f'{cur_memory:.0f} MB')
                if verbose > 1:
                    _dict.update(iter=f'{cur_iter_time:.3f} s',
                                 data=f'{cur_data_time:.3f} s')
                iterator.set_description_str(self.get_str(**_dict, strip=False))
            end = time.time()
        self.synchronize_between_processes()
        total_time = time.time() - start_time
        total_time_str = tqdm_class.format_interval(total_time)

        _dict = {k: v for k, v in self.meters.items()}
        if verbose > 2 and torch.cuda.is_available():
            _dict.update(memory=f'{str(self.memory)} MB')
        if verbose > 1:
            _dict.update(iter=f'{str(self.iter_time)} s',
                         data=f'{str(self.data_time)} s')
        _dict.update(time=total_time_str)
        prints(self.delimiter.join([header, self.get_str(**_dict)]),
               indent=indent)


class AverageMeter:
    r"""Computes and stores the average and current value.

    See Also:
        https://github.com/pytorch/examples/blob/master/imagenet/main.py

    Note:
        It is recommended to use :class:`SmoothedValue` instead.
    """

    def __init__(self, name: str, fmt: str = ':f'):
        self.name: str = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0.
        self.avg = 0.
        self.sum = 0.
        self.count = 0

    def update(self, val: float, n: int = 1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**vars(self))



#!/usr/bin/env python3

from ...abstract import BackdoorDefense
from .adv_train import AdvTrain
from .fine_pruning import FinePruning
from .magnet import MagNet
from .randomized_smooth import RandomizedSmooth
from .recompress import Recompress

__all__ = ['AdvTrain', 'FinePruning', 'MagNet', 'RandomizedSmooth', 'Recompress']

class_dict: dict[str, type[BackdoorDefense]] = {
    'adv_train': AdvTrain,
    'fine_pruning': FinePruning,
    'magnet': MagNet,
    'randomized_smooth': RandomizedSmooth,
    'recompress': Recompress,
}



#!/usr/bin/env python3

from .version import __version__ as internal_version
from torch.torch_version import TorchVersion

from trojanzoo import environ as environ
from trojanzoo import datasets as datasets
from trojanzoo import models as models
from trojanzoo import trainer as trainer

from trojanzoo.utils.module import summary

__all__ = ['summary']
__version__ = TorchVersion(internal_version)



#!/usr/bin/env python3

from trojanvision.datasets.imageset import ImageSet

import torchvision.datasets as datasets
import torchvision.transforms as transforms

import torch
import numpy as np
import os

from typing import Callable, Optional

from trojanvision import __file__ as root_file
root_dir = os.path.dirname(root_file)


class STL10(ImageSet):
    r"""STL10 dataset.
    It inherits :class:`trojanvision.datasets.ImageSet`.

    See Also:
        * torchvision: :any:`torchvision.datasets.STL10`
        * paper: `An Analysis of Single-Layer Networks in Unsupervised Feature Learning`_
        * website: https://cs.stanford.edu/~acoates/stl10/

    Attributes:
        name (str): ``'stl10'``
        num_classes (int): ``10``
        data_shape (list[int]): ``[3, 256, 256]``
        norm_par (dict[str, list[float]]):
            ``{'mean': [0.507, 0.487, 0.441], 'std': [0.267, 0.256, 0.276]}``

    .. _An Analysis of Single-Layer Networks in Unsupervised Feature Learning:
        https://cs.stanford.edu/~acoates/papers/coatesleeng_aistats_2011.pdf
    """

    name: str = 'stl10'
    num_classes: int = 10
    data_shape = [3, 256, 256]

    def __init__(self, norm_par: dict[str, list[int]] = {'mean': [0.507, 0.487, 0.441],
                                                         'std': [0.267, 0.256, 0.276], },
                 **kwargs):
        super().__init__(norm_par=norm_par, **kwargs)

    def initialize(self):
        datasets.STL10(root=self.folder_path, split='train', download=True)
        datasets.STL10(root=self.folder_path, split='test', download=True)
        datasets.STL10(root=self.folder_path, split='unlabeled', download=True)

    def _get_org_dataset(self, mode, **kwargs):
        if mode == 'valid':
            mode = 'test'
        return STL10Dataset(root=self.folder_path, split=mode, **kwargs)

    def get_transform(self, mode: str, normalize: bool = None
                      ) -> transforms.Compose:
        r"""Get dataset transform.

        Args:
            mode (str): The dataset mode (e.g., ``'train' | 'valid' | 'unlabeled' | 'train+unlabeled'``).
            normalize (bool | None):
                Whether to use :any:`torchvision.transforms.Normalize`
                in dataset transform. Defaults to ``self.normalize``.

        Returns:
            torchvision.transforms.Compose: The transform sequence.
        """
        normalize = normalize if normalize is not None else self.normalize

        match mode:
            case 'train' | 'unlabeled' | 'train+unlabeled':
                transform = transforms.Compose([
                    transforms.Resize((272, 272)),
                    transforms.RandomRotation(15,),
                    transforms.RandomCrop(256),
                    transforms.RandomHorizontalFlip(),
                    transforms.PILToTensor(),
                    transforms.ConvertImageDtype(torch.float)])
            case 'valid':
                transform = transforms.Compose([
                    transforms.Resize((256, 256)),
                    transforms.PILToTensor(),
                    transforms.ConvertImageDtype(torch.float)])
            case _:
                raise NotImplementedError(mode)
        if normalize and self.norm_par is not None:
            transform.transforms.append(transforms.Normalize(
                mean=self.norm_par['mean'], std=self.norm_par['std']))
        return transform


class STL10Dataset(datasets.STL10):
    train_list = [
        ["train_X.bin", "918c2871b30a85fa023e0c44e0bee87f"],
        ["train_y.bin", "5a34089d4802c674881badbb80307741"],
        ["unlabeled_X.bin", "5242ba1fed5e4be9e1e742405eb56ca4"],
        ["unlabeled_y.bin", None],
    ]

    def __init__(
        self,
        root: str,
        split: str = "train",
        folds: Optional[int] = None,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ) -> None:
        super().__init__(root, split=split, folds=folds,
                         transform=transform, target_transform=target_transform,
                         download=download)
        match self.split:
            case "train+unlabeled":
                labels = self.read_labels()
                self.labels = np.concatenate((self.labels[:-labels.shape[0]], labels))
            case "unlabeled":
                self.labels = self.read_labels()

    @staticmethod
    def read_labels() -> np.ndarray:
        path_to_labels = os.path.normpath(os.path.join(root_dir, 'data', 'stl10', 'unlabeled_y.bin'))
        with open(path_to_labels, "rb") as f:
            labels = np.fromfile(f, dtype=np.uint8) - 1  # 0-based
        return labels



#!/usr/bin/env python3

from ...abstract import BackdoorAttack

from .badnet import BadNet
from .trojannn import TrojanNN
from .latent_backdoor import LatentBackdoor
from .imc import IMC
from .trojannet import TrojanNet

__all__ = ['BadNet', 'TrojanNN', 'LatentBackdoor', 'IMC', 'TrojanNet']

class_dict: dict[str, type[BackdoorAttack]] = {
    'badnet': BadNet,
    'trojannn': TrojanNN,
    'latent_backdoor': LatentBackdoor,
    'imc': IMC,
    'trojannet': TrojanNet,
}



#!/usr/bin/env python3

from trojanvision.models.imagemodel import ImageModel

from .bit import BiT
from .dla import DLA
from .dpn import DPN
from .net import Net

__all__ = ['BiT', 'DLA', 'DPN', 'Net']

class_dict: dict[str, type[ImageModel]] = {
    'bit': BiT,
    'dla': DLA,
    'dpn': DPN,
    'net': Net,
}



#!/usr/bin/env python3

r"""
CUDA_VISIBLE_DEVICES=0 python examples/validate.py --color --verbose 1 --dataset cifar10 --pretrained --model resnet18_comp
"""  # noqa: E501

import trojanvision
import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    trojanvision.environ.add_argument(parser)
    trojanvision.datasets.add_argument(parser)
    trojanvision.models.add_argument(parser)
    kwargs = vars(parser.parse_args())

    env = trojanvision.environ.create(**kwargs)
    dataset = trojanvision.datasets.create(**kwargs)
    model = trojanvision.models.create(dataset=dataset, **kwargs)

    if env['verbose']:
        trojanvision.summary(env=env, dataset=dataset, model=model)
    acc, loss = model._validate()
